# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""cubicweb.server.sources.ldapfeed unit and functional tests

Those tests expect to have slapd, python-ldap3 and ldapscripts packages installed.
"""

from __future__ import print_function

import os
import sys
import shutil
import time
import subprocess
import tempfile
import unittest
from os.path import join

from six import string_types
from six.moves import range

from cubicweb import AuthenticationError
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.devtools.httptest import get_available_port


CONFIG_LDAPFEED = u'''
user-base-dn=ou=People,dc=cubicweb,dc=test
group-base-dn=ou=Group,dc=cubicweb,dc=test
user-attrs-map=uid=login,mail=email,userPassword=upassword
group-attrs-map=cn=name,memberUid=member
'''

URL = None


def create_slapd_configuration(cls):
    global URL
    slapddir = tempfile.mkdtemp('cw-unittest-ldap')
    config = cls.config
    slapdconf = join(config.apphome, "slapd.conf")
    confin = open(join(config.apphome, "slapd.conf.in")).read()
    confstream = open(slapdconf, 'w')
    confstream.write(confin % {'apphome': config.apphome, 'testdir': slapddir})
    confstream.close()
    # fill ldap server with some data
    ldiffile = join(config.apphome, "ldap_test.ldif")
    config.info('Initing ldap database')
    cmdline = ['/usr/sbin/slapadd', '-f', slapdconf, '-l', ldiffile, '-c']
    PIPE = subprocess.PIPE
    slapproc = subprocess.Popen(cmdline, stdout=PIPE, stderr=PIPE)
    stdout, stderr = slapproc.communicate()
    if slapproc.returncode:
        print('slapadd returned with status: %s'
              % slapproc.returncode, file=sys.stderr)
        sys.stdout.write(stdout)
        sys.stderr.write(stderr)

    # ldapuri = 'ldapi://' + join(basedir, "ldapi").replace('/', '%2f')
    port = get_available_port(range(9000, 9100))
    host = 'localhost:%s' % port
    ldapuri = 'ldap://%s' % host
    cmdline = ["/usr/sbin/slapd", "-f", slapdconf, "-h", ldapuri, "-d", "0"]
    config.info('Starting slapd:', ' '.join(cmdline))
    PIPE = subprocess.PIPE
    cls.slapd_process = subprocess.Popen(cmdline, stdout=PIPE, stderr=PIPE)
    time.sleep(0.2)
    if cls.slapd_process.poll() is None:
        config.info('slapd started with pid %s', cls.slapd_process.pid)
    else:
        raise EnvironmentError('Cannot start slapd with cmdline="%s" (from directory "%s")' %
                               (" ".join(cmdline), os.getcwd()))
    URL = u'ldap://%s' % host
    return slapddir


def terminate_slapd(cls):
    config = cls.config
    if cls.slapd_process and cls.slapd_process.returncode is None:
        config.info('terminating slapd')
        if hasattr(cls.slapd_process, 'terminate'):
            cls.slapd_process.terminate()
        else:
            import signal
            os.kill(cls.slapd_process.pid, signal.SIGTERM)
        stdout, stderr = cls.slapd_process.communicate()
        if cls.slapd_process.returncode:
            print('slapd returned with status: %s'
                  % cls.slapd_process.returncode, file=sys.stderr)
            sys.stdout.write(stdout)
            sys.stderr.write(stderr)
        config.info('DONE')
    try:
        shutil.rmtree(cls._tmpdir)
    except:
        pass


class LDAPFeedTestBase(CubicWebTC):
    test_db_id = 'ldap-feed'
    loglevel = 'ERROR'

    @classmethod
    def setUpClass(cls):
        super(LDAPFeedTestBase, cls).setUpClass()
        cls.init_slapd()

    @classmethod
    def tearDownClass(cls):
        cls.terminate_slapd()

    @classmethod
    def init_slapd(cls):
        for path in ('/usr/sbin/slapd',
                     '/usr/sbin/slapadd',
                     '/usr/bin/ldapmodify'):
            if not os.path.exists(path):
                raise unittest.SkipTest('%s not found' % path)
        from cubicweb.cwctl import init_cmdline_log_threshold
        init_cmdline_log_threshold(cls.config, cls.loglevel)
        cls._tmpdir = create_slapd_configuration(cls)

    @classmethod
    def terminate_slapd(cls):
        terminate_slapd(cls)

    @classmethod
    def pre_setup_database(cls, cnx, config):
        cnx.create_entity('CWSource', name=u'ldap', type=u'ldapfeed', parser=u'ldapfeed',
                          url=URL, config=CONFIG_LDAPFEED)

        cnx.commit()
        return cls.pull(cnx)

    @classmethod
    def pull(self, cnx):
        lfsource = cnx.repo.sources_by_uri['ldap']
        stats = lfsource.pull_data(cnx, force=True, raise_on_error=True)
        cnx.commit()
        return stats

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('DELETE Any E WHERE E cw_source S, S name "ldap"')
            cnx.execute('SET S config %(conf)s, S url %(url)s '
                        'WHERE S is CWSource, S name "ldap"',
                        {"conf": CONFIG_LDAPFEED, 'url': URL})
            cnx.commit()
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)

    def add_ldap_entry(self, dn, mods):
        """
        add an LDAP entity
        """
        modcmd = ['dn: %s' % dn, 'changetype: add']
        for key, values in mods.items():
            if isinstance(values, string_types):
                values = [values]
            for value in values:
                modcmd.append('%s: %s' % (key, value))
        self._ldapmodify(modcmd)

    def delete_ldap_entry(self, dn):
        """
        delete an LDAP entity
        """
        modcmd = ['dn: %s' % dn, 'changetype: delete']
        self._ldapmodify(modcmd)

    def update_ldap_entry(self, dn, mods):
        """
        modify one or more attributes of an LDAP entity
        """
        modcmd = ['dn: %s' % dn, 'changetype: modify']
        for (kind, key), values in mods.items():
            modcmd.append('%s: %s' % (kind, key))
            if isinstance(values, string_types):
                values = [values]
            for value in values:
                modcmd.append('%s: %s' % (key, value))
            modcmd.append('-')
        self._ldapmodify(modcmd)

    def _ldapmodify(self, modcmd):
        uri = self.repo.sources_by_uri['ldap'].urls[0]
        updatecmd = ['ldapmodify', '-H', uri, '-v', '-x', '-D',
                     'cn=admin,dc=cubicweb,dc=test', '-w', 'cw']
        PIPE = subprocess.PIPE
        p = subprocess.Popen(updatecmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        p.stdin.write('\n'.join(modcmd).encode('ascii'))
        p.stdin.close()
        if p.wait():
            raise RuntimeError("ldap update failed: %s" % ('\n'.join(p.stderr.readlines())))


class CheckWrongGroup(LDAPFeedTestBase):
    """
    A testcase for situations where the default group for CWUser
    created from LDAP is wrongly configured.
    """

    def test_wrong_group(self):
        with self.admin_access.repo_cnx() as cnx:
            source = cnx.execute('CWSource S WHERE S type="ldapfeed"').get_entity(0, 0)
            config = source.repo_source.check_config(source)
            # inject a bogus group here, along with at least a valid one
            config['user-default-group'] = ('thisgroupdoesnotexists', 'users')
            source.repo_source.update_config(source, config)
            cnx.commit()
            # here we emitted an error log entry
            source.repo_source.pull_data(cnx, force=True, raise_on_error=True)
            cnx.commit()


class LDAPFeedUserTC(LDAPFeedTestBase):
    """
    A testcase for CWUser support in ldapfeed (basic tests and authentication).
    """

    def assertMetadata(self, entity):
        self.assertTrue(entity.creation_date)
        self.assertTrue(entity.modification_date)

    def test_authenticate(self):
        source = self.repo.sources_by_uri['ldap']
        with self.admin_access.repo_cnx() as cnx:
            # ensure we won't be logged against
            self.assertRaises(AuthenticationError,
                              source.authenticate, cnx, 'toto', 'toto')
            self.assertRaises(AuthenticationError,
                              source.authenticate, cnx, 'syt', 'toto')
            self.assertTrue(source.authenticate(cnx, 'syt', 'syt'))

    def test_base(self):
        with self.admin_access.repo_cnx() as cnx:
            # check a known one
            rset = cnx.execute('CWUser X WHERE X login %(login)s', {'login': 'syt'})
            e = rset.get_entity(0, 0)
            self.assertEqual(e.login, 'syt')
            e.complete()
            self.assertMetadata(e)
            self.assertEqual(e.firstname, None)
            self.assertEqual(e.surname, None)
            self.assertIn('users', set(g.name for g in e.in_group))
            self.assertEqual(e.owned_by[0].login, 'syt')
            self.assertEqual(e.created_by, ())
            addresses = [pe.address for pe in e.use_email]
            addresses.sort()
            self.assertEqual(['sylvain.thenault@logilab.fr', 'syt@logilab.fr'],
                             addresses)
            self.assertIn(e.primary_email[0].address, ['sylvain.thenault@logilab.fr',
                                                       'syt@logilab.fr'])
            # email content should be indexed on the user
            rset = cnx.execute('CWUser X WHERE X has_text "thenault"')
            self.assertEqual(rset.rows, [[e.eid]])

    def test_copy_to_system_source(self):
        "make sure we can 'convert' an LDAP user into a system one"
        with self.admin_access.repo_cnx() as cnx:
            source = self.repo.sources_by_uri['ldap']
            eid = cnx.execute('CWUser X WHERE X login %(login)s', {'login': 'syt'})[0][0]
            cnx.execute('SET X cw_source S WHERE X eid %(x)s, S name "system"', {'x': eid})
            cnx.commit()
            rset = cnx.execute('CWUser X WHERE X login %(login)s', {'login': 'syt'})
            self.assertEqual(len(rset), 1)
            e = rset.get_entity(0, 0)
            self.assertEqual(e.eid, eid)
            self.assertEqual(e.cw_source[0].name, 'system')
            self.assertTrue(e.creation_date)
            self.assertTrue(e.modification_date)
            source.pull_data(cnx, raise_on_error=True)
            rset = cnx.execute('CWUser X WHERE X login %(login)s', {'login': 'syt'})
            self.assertEqual(len(rset), 1)
            self.assertTrue(self.repo.system_source.authenticate(cnx, 'syt', password='syt'))
            # make sure the pull from ldap have not "reverted" user as a ldap-feed user
            # and that the password stored in the system source is not empty or so
            user = cnx.execute('CWUser U WHERE U login "syt"').get_entity(0, 0)
            user.cw_clear_all_caches()
            self.assertEqual(user.cw_source[0].name, 'system')
            cu = cnx.system_sql("SELECT cw_upassword FROM cw_cwuser WHERE cw_login='syt';")
            pwd = cu.fetchall()[0][0]
            self.assertIsNotNone(pwd)
            self.assertTrue(str(pwd))


class LDAPGeneratePwdTC(LDAPFeedTestBase):
    """
    A testcase for password generation on CWUser when none is imported
    """

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            lfsource = cnx.repo.sources_by_uri['ldap']
            del lfsource.user_attrs['userPassword']
        super(LDAPGeneratePwdTC, self).setup_database()

    def test_no_password(self):
        with self.admin_access.repo_cnx() as cnx:
            cu = cnx.system_sql("SELECT cw_upassword FROM cw_cwuser WHERE cw_login='syt';")
            pwd = cu.fetchall()[0][0]
            self.assertTrue(pwd)


class LDAPFeedUserDeletionTC(LDAPFeedTestBase):
    """
    A testcase for situations where users are deleted from or
    unavailable in the LDAP database.
    """

    def test_a_filter_inactivate(self):
        """ filtered out people should be deactivated, unable to authenticate """
        repo_source = self.repo.sources_by_uri['ldap']
        with self.admin_access.repo_cnx() as cnx:
            source = cnx.execute('CWSource S WHERE S type="ldapfeed"').get_entity(0, 0)
            config = repo_source.check_config(source)
            # filter with adim's phone number
            config['user-filter'] = u'(%s=%s)' % ('telephoneNumber', '109')
            repo_source.update_config(source, config)
            cnx.commit()
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
            self.assertRaises(AuthenticationError,
                              repo_source.authenticate, cnx, 'syt', 'syt')
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(cnx.execute('Any N WHERE U login "syt", '
                                         'U in_state S, S name N').rows[0][0],
                             'deactivated')
            self.assertEqual(cnx.execute('Any N WHERE U login "adim", '
                                         'U in_state S, S name N').rows[0][0],
                             'activated')
            # unfilter, syt should be activated again
            config['user-filter'] = u''
            repo_source.update_config(source, config)
            cnx.commit()
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(cnx.execute('Any N WHERE U login "syt", '
                                         'U in_state S, S name N').rows[0][0],
                             'activated')
            self.assertEqual(cnx.execute('Any N WHERE U login "adim", '
                                         'U in_state S, S name N').rows[0][0],
                             'activated')

    def test_delete(self):
        """ delete syt, pull, check deactivation, repull,
        read syt, pull, check activation
        """
        self.delete_ldap_entry('uid=syt,ou=People,dc=cubicweb,dc=test')
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
            source = self.repo.sources_by_uri['ldap']
            self.assertRaises(AuthenticationError,
                              source.authenticate, cnx, 'syt', 'syt')
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(cnx.execute('Any N WHERE U login "syt", '
                                         'U in_state S, S name N').rows[0][0],
                             'deactivated')
        with self.repo.internal_cnx() as cnx:
            # check that it doesn't choke
            self.pull(cnx)
        # reinsert syt
        self.add_ldap_entry('uid=syt,ou=People,dc=cubicweb,dc=test',
                            {'objectClass': ['OpenLDAPperson', 'posixAccount', 'top',
                                             'shadowAccount'],
                             'cn': 'Sylvain Thenault',
                             'sn': 'Thenault',
                             'gidNumber': '1004',
                             'uid': 'syt',
                             'homeDirectory': '/home/syt',
                             'shadowFlag': '134538764',
                             'uidNumber': '1004',
                             'givenName': 'Sylvain',
                             'telephoneNumber': '106',
                             'displayName': 'sthenault',
                             'gecos': 'Sylvain Thenault',
                             'mail': ['sylvain.thenault@logilab.fr', 'syt@logilab.fr'],
                             'userPassword': 'syt',
                             })
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(cnx.execute('Any N WHERE U login "syt", '
                                         'U in_state S, S name N').rows[0][0],
                             'activated')

    def test_reactivate_deleted(self):
        # test reactivating BY HAND the user isn't enough to
        # authenticate, as the native source refuse to authenticate
        # user from other sources
        repo_source = self.repo.sources_by_uri['ldap']
        self.delete_ldap_entry('uid=syt,ou=People,dc=cubicweb,dc=test')
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            # reactivate user (which source is still ldap-feed)
            user = cnx.execute('CWUser U WHERE U login "syt"').get_entity(0, 0)
            user.cw_adapt_to('IWorkflowable').fire_transition('activate')
            cnx.commit()
            self.assertRaises(AuthenticationError,
                              repo_source.authenticate, cnx, 'syt', 'syt')

            # ok now let's try to make it a system user
            cnx.execute('SET X cw_source S WHERE X eid %(x)s, S name "system"', {'x': user.eid})
            cnx.commit()
            # and that we can now authenticate again
            self.assertRaises(AuthenticationError,
                              repo_source.authenticate, cnx, 'syt', 'toto')
            self.assertTrue(self.repo.authenticate_user(cnx, 'syt', password='syt'))


class LDAPFeedGroupTC(LDAPFeedTestBase):
    """
    A testcase for group support in ldapfeed.
    """

    def test_groups_exist(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute('CWGroup X WHERE X name "dir"')
            self.assertEqual(len(rset), 1)

            rset = cnx.execute('CWGroup X WHERE X cw_source S, S name "ldap"')
            self.assertEqual(len(rset), 2)

    def test_group_deleted(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute('CWGroup X WHERE X name "dir"')
            self.assertEqual(len(rset), 1)

    def test_in_group(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute('CWGroup X WHERE X name %(name)s', {'name': 'dir'})
            dirgroup = rset.get_entity(0, 0)
            self.assertEqual(set(['syt', 'adim']),
                             set([u.login for u in dirgroup.reverse_in_group]))
            rset = cnx.execute('CWGroup X WHERE X name %(name)s', {'name': 'logilab'})
            logilabgroup = rset.get_entity(0, 0)
            self.assertEqual(set(['adim']),
                             set([u.login for u in logilabgroup.reverse_in_group]))

    def test_group_member_added(self):
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute('Any L WHERE U in_group G, G name %(name)s, U login L',
                               {'name': 'logilab'})
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], 'adim')

        try:
            self.update_ldap_entry('cn=logilab,ou=Group,dc=cubicweb,dc=test',
                                   {('add', 'memberUid'): ['syt']})
            with self.repo.internal_cnx() as cnx:
                self.pull(cnx)

            with self.admin_access.repo_cnx() as cnx:
                rset = cnx.execute('Any L WHERE U in_group G, G name %(name)s, U login L',
                                   {'name': 'logilab'})
                self.assertEqual(len(rset), 2)
                members = set([u[0] for u in rset])
                self.assertEqual(set(['adim', 'syt']), members)

        finally:
            # back to normal ldap setup
            self.terminate_slapd()
            self.init_slapd()

    def test_group_member_deleted(self):
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)  # ensure we are sync'ed
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute('Any L WHERE U in_group G, G name %(name)s, U login L',
                               {'name': 'logilab'})
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], 'adim')

        try:
            self.update_ldap_entry('cn=logilab,ou=Group,dc=cubicweb,dc=test',
                                   {('delete', 'memberUid'): ['adim']})
            with self.repo.internal_cnx() as cnx:
                self.pull(cnx)

            with self.admin_access.repo_cnx() as cnx:
                rset = cnx.execute('Any L WHERE U in_group G, G name %(name)s, U login L',
                                   {'name': 'logilab'})
                self.assertEqual(len(rset), 0, rset.rows)
        finally:
            # back to normal ldap setup
            self.terminate_slapd()
            self.init_slapd()


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# -*- coding: utf-8 -*-
# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
from six import binary_type

from cubicweb.devtools.testlib import CubicWebTC


class JsonViewsTC(CubicWebTC):
    anonymize = True
    res_jsonp_data = b'[["guests", 1]]'

    def setUp(self):
        super(JsonViewsTC, self).setUp()
        self.config.global_set_option('anonymize-jsonp-queries', self.anonymize)

    def test_json_rsetexport(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
                'Any GN,COUNT(X) GROUPBY GN ORDERBY GN WHERE X in_group G, G name GN')
            data = self.view('jsonexport', rset, req=req)
            self.assertEqual(req.headers_out.getRawHeaders('content-type'), ['application/json'])
            self.assertListEqual(data, [["guests", 1], ["managers", 1]])

    def test_json_rsetexport_empty_rset(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(u'Any X WHERE X is CWUser, X login "foobarbaz"')
            data = self.view('jsonexport', rset, req=req)
            self.assertEqual(req.headers_out.getRawHeaders('content-type'), ['application/json'])
            self.assertListEqual(data, [])

    def test_json_rsetexport_with_jsonp(self):
        with self.admin_access.web_request() as req:
            req.form.update({'callback': u'foo',
                             'rql': u'Any GN,COUNT(X) GROUPBY GN ORDERBY GN '
                             'WHERE X in_group G, G name GN'})
            data = self.ctrl_publish(req, ctrl='jsonp')
            self.assertIsInstance(data, binary_type)
            self.assertEqual(req.headers_out.getRawHeaders('content-type'),
                             ['application/javascript'])
            # because jsonp anonymizes data, only 'guests' group should be found
            self.assertEqual(data, b'foo(' + self.res_jsonp_data + b')')

    def test_json_rsetexport_with_jsonp_and_bad_vid(self):
        with self.admin_access.web_request() as req:
            req.form.update({'callback': 'foo',
                             # "vid" parameter should be ignored by jsonp controller
                             'vid': 'table',
                             'rql': 'Any GN,COUNT(X) GROUPBY GN ORDERBY GN '
                             'WHERE X in_group G, G name GN'})
            data = self.ctrl_publish(req, ctrl='jsonp')
            self.assertEqual(req.headers_out.getRawHeaders('content-type'),
                             ['application/javascript'])
            # result should be plain json, not the table view
            self.assertEqual(data, b'foo(' + self.res_jsonp_data + b')')

    def test_json_ersetexport(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any G ORDERBY GN WHERE G is CWGroup, G name GN')
            data = self.view('ejsonexport', rset, req=req)
            self.assertEqual(req.headers_out.getRawHeaders('content-type'), ['application/json'])
            self.assertEqual(data[0]['name'], 'guests')
            self.assertEqual(data[1]['name'], 'managers')

            rset = req.execute(u'Any G WHERE G is CWGroup, G name "foo"')
            data = self.view('ejsonexport', rset, req=req)
            self.assertEqual(req.headers_out.getRawHeaders('content-type'), ['application/json'])
            self.assertEqual(data, [])


class NotAnonymousJsonViewsTC(JsonViewsTC):
    anonymize = False
    res_jsonp_data = b'[["guests", 1], ["managers", 1]]'


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Security hooks: check permissions to add/delete/update entities according to
the connected user
"""

from logilab.common.registry import objectify_predicate

from yams import buildobjs

from cubicweb import Unauthorized
from cubicweb.server import BEFORE_ADD_RELATIONS, ON_COMMIT_ADD_RELATIONS, hook


def check_entity_attributes(cnx, entity, action, editedattrs=None):
    eid = entity.eid
    eschema = entity.e_schema
    # ._cw_skip_security_attributes is there to bypass security for attributes
    # set by hooks by modifying the entity's dictionary
    if editedattrs is None:
        editedattrs = entity.cw_edited
    dontcheck = editedattrs.skip_security
    etypechecked = False
    for attr in editedattrs:
        if attr in dontcheck:
            continue
        rdef = eschema.rdef(attr, takefirst=True)
        if rdef.final: # non final relation are checked by standard hooks
            perms = rdef.permissions.get(action)
            # comparison below works because the default update perm is:
            #
            #  ('managers', ERQLExpression(Any X WHERE U has_update_permission X,
            #                              X eid %(x)s, U eid %(u)s))
            #
            # is deserialized in this order (groups first), and ERQLExpression
            # implements comparison by rql expression.
            if perms == buildobjs.DEFAULT_ATTRPERMS[action]:
                # The default rule is to delegate to the entity
                # rule. This needs to be checked only once.
                if not etypechecked:
                    entity.cw_check_perm(action)
                    etypechecked = True
                continue
            if perms == ():
                # That means an immutable attribute; as an optimization, avoid
                # going through check_perm.
                raise Unauthorized(action, str(rdef))
            rdef.check_perm(cnx, action, eid=eid)

    if action == 'add' and not etypechecked:
        # think about cnx.create_entity('Foo')
        # the standard metadata were inserted by a hook
        # with a bypass ... we conceptually need to check
        # the eid attribute at *creation* time
        entity.cw_check_perm(action)


class CheckEntityPermissionOp(hook.DataOperationMixIn, hook.LateOperation):
    def precommit_event(self):
        cnx = self.cnx
        for eid, action, edited in self.get_data():
            entity = cnx.entity_from_eid(eid)
            check_entity_attributes(cnx, entity, action, edited)


class CheckRelationPermissionOp(hook.DataOperationMixIn, hook.LateOperation):
    def precommit_event(self):
        cnx = self.cnx
        for action, rschema, eidfrom, eidto in self.get_data():
            rdef = rschema.rdef(cnx.entity_type(eidfrom),
                                cnx.entity_type(eidto))
            rdef.check_perm(cnx, action, fromeid=eidfrom, toeid=eidto)


@objectify_predicate
def write_security_enabled(cls, req, **kwargs):
    if req is None or not req.write_security:
        return 0
    return 1

class SecurityHook(hook.Hook):
    __abstract__ = True
    category = 'security'
    __select__ = hook.Hook.__select__ & write_security_enabled()


class AfterAddEntitySecurityHook(SecurityHook):
    __regid__ = 'securityafteraddentity'
    events = ('after_add_entity',)

    def __call__(self):
        CheckEntityPermissionOp.get_instance(self._cw).add_data(
            (self.entity.eid, 'add', self.entity.cw_edited) )


class AfterUpdateEntitySecurityHook(SecurityHook):
    __regid__ = 'securityafterupdateentity'
    events = ('after_update_entity',)

    def __call__(self):
        # save back editedattrs in case the entity is reedited later in the
        # same transaction, which will lead to cw_edited being
        # overwritten
        action = 'add' if self._cw.added_in_transaction(self.entity.eid) else 'update'
        CheckEntityPermissionOp.get_instance(self._cw).add_data(
            (self.entity.eid, action, self.entity.cw_edited) )


class BeforeDelEntitySecurityHook(SecurityHook):
    __regid__ = 'securitybeforedelentity'
    events = ('before_delete_entity',)

    def __call__(self):
        self.entity.cw_check_perm('delete')


def skip_inlined_relation_security(cnx, rschema, eid):
    """return True if security for the given inlined relation should be skipped,
    in case where the relation has been set through modification of
    `entity.cw_edited` in a hook
    """
    assert rschema.inlined
    try:
        entity = cnx.entity_cache(eid)
    except KeyError:
        return False
    edited = getattr(entity, 'cw_edited', None)
    if edited is None:
        return False
    return rschema.type in edited.skip_security


class BeforeAddRelationSecurityHook(SecurityHook):
    __regid__ = 'securitybeforeaddrelation'
    events = ('before_add_relation',)

    def __call__(self):
        if self.rtype in BEFORE_ADD_RELATIONS:
            nocheck = self._cw.transaction_data.get('skip-security', ())
            if (self.eidfrom, self.rtype, self.eidto) in nocheck:
                return
            rschema = self._cw.repo.schema[self.rtype]
            if rschema.inlined and skip_inlined_relation_security(
                    self._cw, rschema, self.eidfrom):
                return
            rdef = rschema.rdef(self._cw.entity_type(self.eidfrom),
                                self._cw.entity_type(self.eidto))
            rdef.check_perm(self._cw, 'add', fromeid=self.eidfrom, toeid=self.eidto)


class AfterAddRelationSecurityHook(SecurityHook):
    __regid__ = 'securityafteraddrelation'
    events = ('after_add_relation',)

    def __call__(self):
        if self.rtype not in BEFORE_ADD_RELATIONS:
            nocheck = self._cw.transaction_data.get('skip-security', ())
            if (self.eidfrom, self.rtype, self.eidto) in nocheck:
                return
            rschema = self._cw.repo.schema[self.rtype]
            if rschema.inlined and skip_inlined_relation_security(
                    self._cw, rschema, self.eidfrom):
                return
            if self.rtype in ON_COMMIT_ADD_RELATIONS:
                CheckRelationPermissionOp.get_instance(self._cw).add_data(
                    ('add', rschema, self.eidfrom, self.eidto) )
            else:
                rdef = rschema.rdef(self._cw.entity_type(self.eidfrom),
                                    self._cw.entity_type(self.eidto))
                rdef.check_perm(self._cw, 'add', fromeid=self.eidfrom, toeid=self.eidto)


class BeforeDeleteRelationSecurityHook(SecurityHook):
    __regid__ = 'securitybeforedelrelation'
    events = ('before_delete_relation',)

    def __call__(self):
        nocheck = self._cw.transaction_data.get('skip-security', ())
        if (self.eidfrom, self.rtype, self.eidto) in nocheck:
            return
        rschema = self._cw.repo.schema[self.rtype]
        if rschema.inlined and skip_inlined_relation_security(
                self._cw, rschema, self.eidfrom):
            return
        rdef = rschema.rdef(self._cw.entity_type(self.eidfrom),
                            self._cw.entity_type(self.eidto))
        rdef.check_perm(self._cw, 'delete', fromeid=self.eidfrom, toeid=self.eidto)



# copyright 2003-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.server.migractions"""

import os
import os.path as osp
import sys
from datetime import date
from contextlib import contextmanager
import tempfile

from logilab.common.testlib import unittest_main, Tags, tag
from logilab.common import tempattr

from yams.constraints import UniqueConstraint

from cubicweb import (ConfigurationError, ValidationError,
                      ExecutionError, Binary)
from cubicweb.devtools import startpgcluster, stoppgcluster
from cubicweb.devtools.testlib import CubicWebTC, TemporaryDirectory
from cubicweb.server.sqlutils import SQL_PREFIX
from cubicweb.server.migractions import ServerMigrationHelper
from cubicweb.server.sources import storages
from cubicweb.server.schema2sql import build_index_name

import cubicweb.devtools


HERE = osp.dirname(osp.abspath(__file__))
migrschema = None


def setUpModule():
    startpgcluster(__file__)


def tearDownModule(*args):
    global migrschema
    del migrschema
    if hasattr(MigrationCommandsTC, 'origschema'):
        del MigrationCommandsTC.origschema
    if hasattr(MigrationCommandsComputedTC, 'origschema'):
        del MigrationCommandsComputedTC.origschema
    stoppgcluster(__file__)


class MigrationConfig(cubicweb.devtools.TestServerConfiguration):
    default_sources = cubicweb.devtools.DEFAULT_PSQL_SOURCES
    CUBES_PATH = cubicweb.devtools.TestServerConfiguration.CUBES_PATH + [
        osp.join(HERE, 'data-migractions', 'cubes')]


class MigrationTC(CubicWebTC):

    appid = 'data-migractions'

    configcls = MigrationConfig

    tags = CubicWebTC.tags | Tags(('server', 'migration', 'migractions'))

    def _init_repo(self):
        super(MigrationTC, self)._init_repo()
        # we have to read schema from the database to get eid for schema entities
        self.repo.set_schema(self.repo.deserialize_schema(), resetvreg=False)
        # hack to read the schema from data/migrschema

        @contextmanager
        def temp_app(config, appid, apphome):
            old = config.apphome, config.appid
            sys.path.remove(old[0])
            sys.path.insert(0, apphome)
            config._apphome, config.appid = apphome, appid
            try:
                yield config
            finally:
                sys.path.remove(apphome)
                sys.path.insert(0, old[0])
                config._apphome, config.appid = old

        appid = osp.join(self.appid, 'migratedapp')
        apphome = osp.join(HERE, appid)
        with temp_app(self.config, appid, apphome) as config:
            global migrschema
            migrschema = config.load_schema()

    def setUp(self):
        self.configcls.cls_adjust_sys_path()
        super(MigrationTC, self).setUp()

    def tearDown(self):
        super(MigrationTC, self).tearDown()
        self.repo.vreg['etypes'].clear_caches()

    @contextmanager
    def mh(self):
        with self.admin_access.repo_cnx() as cnx:
            yield cnx, ServerMigrationHelper(self.repo.config, migrschema,
                                             repo=self.repo, cnx=cnx,
                                             interactive=False)

    def table_sql(self, mh, tablename):
        result = mh.sqlexec("SELECT table_name FROM information_schema.tables WHERE LOWER(table_name)=%(table)s",
                            {'table': tablename.lower()})
        if result:
            return result[0][0]
        return None # no such table

    def table_schema(self, mh, tablename):
        result = mh.sqlexec("SELECT column_name, data_type, character_maximum_length FROM information_schema.columns "
                            "WHERE LOWER(table_name) = %(table)s", {'table': tablename.lower()})
        assert result, 'no table %s' % tablename
        return dict((x[0], (x[1], x[2])) for x in result)

    def table_constraints(self, mh, tablename):
        result = mh.sqlexec(
            "SELECT DISTINCT constraint_name FROM information_schema.constraint_column_usage "
            "WHERE LOWER(table_name) = '%(table)s' AND constraint_name LIKE 'cstr%%'"
            % {'table': tablename.lower()})
        assert result, 'no table %s' % tablename
        return set(x[0] for x in result)


class MigrationCommandsTC(MigrationTC):

    def _init_repo(self):
        super(MigrationCommandsTC, self)._init_repo()
        assert 'Folder' in migrschema

    def test_add_attribute_bool(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn('yesno', self.schema)
            cnx.create_entity('Note')
            cnx.commit()
            mh.cmd_add_attribute('Note', 'yesno')
            self.assertIn('yesno', self.schema)
            self.assertEqual(self.schema['yesno'].subjects(), ('Note',))
            self.assertEqual(self.schema['yesno'].objects(), ('Boolean',))
            self.assertEqual(self.schema['Note'].default('yesno'), False)
            # test default value set on existing entities
            note = cnx.execute('Note X').get_entity(0, 0)
            self.assertEqual(note.yesno, False)
            # test default value set for next entities
            self.assertEqual(cnx.create_entity('Note').yesno, False)

    def test_add_attribute_int(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn('whatever', self.schema)
            cnx.create_entity('Note')
            cnx.commit()
            orderdict = dict(mh.rqlexec('Any RTN, O WHERE X name "Note", RDEF from_entity X, '
                                        'RDEF relation_type RT, RDEF ordernum O, RT name RTN'))
            mh.cmd_add_attribute('Note', 'whatever')
            self.assertIn('whatever', self.schema)
            self.assertEqual(self.schema['whatever'].subjects(), ('Note',))
            self.assertEqual(self.schema['whatever'].objects(), ('Int',))
            self.assertEqual(self.schema['Note'].default('whatever'), 0)
            # test default value set on existing entities
            note = cnx.execute('Note X').get_entity(0, 0)
            self.assertIsInstance(note.whatever, int)
            self.assertEqual(note.whatever, 0)
            # test default value set for next entities
            self.assertEqual(cnx.create_entity('Note').whatever, 0)
            # test attribute order
            orderdict2 = dict(mh.rqlexec('Any RTN, O WHERE X name "Note", RDEF from_entity X, '
                                         'RDEF relation_type RT, RDEF ordernum O, RT name RTN'))
            whateverorder = migrschema['whatever'].rdef('Note', 'Int').order
            for k, v in orderdict.items():
                if v >= whateverorder:
                    orderdict[k] = v+1
            orderdict['whatever'] = whateverorder
            self.assertDictEqual(orderdict, orderdict2)
            #self.assertEqual([r.type for r in self.schema['Note'].ordered_relations()],
            #                  ['modification_date', 'creation_date', 'owned_by',
            #                   'eid', 'ecrit_par', 'inline1', 'date', 'type',
            #                   'whatever', 'date', 'in_basket'])
            # NB: commit instead of rollback make following test fail with py2.5
            #     this sounds like a pysqlite/2.5 bug (the same eid is affected to
            #     two different entities)

    def test_add_attribute_varchar(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn('whatever', self.schema)
            cnx.create_entity('Note')
            cnx.commit()
            self.assertNotIn('shortpara', self.schema)
            mh.cmd_add_attribute('Note', 'shortpara')
            self.assertIn('shortpara', self.schema)
            self.assertEqual(self.schema['shortpara'].subjects(), ('Note', ))
            self.assertEqual(self.schema['shortpara'].objects(), ('String', ))
            # test created column is actually a varchar(64)
            fields = self.table_schema(mh, '%sNote' % SQL_PREFIX)
            self.assertEqual(fields['%sshortpara' % SQL_PREFIX], ('character varying', 11))
            # test default value set on existing entities
            self.assertEqual(cnx.execute('Note X').get_entity(0, 0).shortpara, 'hop')
            # test default value set for next entities
            self.assertEqual(cnx.create_entity('Note', shortpara=u'hop hop').shortpara, u'hop hop')
            # serialized constraint added
            constraints = self.table_constraints(mh, 'cw_Personne')
            self.assertEqual(len(constraints), 1, constraints)

    def test_add_attribute_unique(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn('unique_id', self.schema)
            mh.cmd_add_attribute('Note', 'unique_id')
            # test unique index creation
            dbh = self.repo.system_source.dbhelper
            self.assertTrue(dbh.index_exists(cnx.cnxset.cu, 'cw_Note', 'cw_unique_id', unique=True))

    def test_add_datetime_with_default_value_attribute(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn('mydate', self.schema)
            self.assertNotIn('oldstyledefaultdate', self.schema)
            self.assertNotIn('newstyledefaultdate', self.schema)
            mh.cmd_add_attribute('Note', 'mydate')
            mh.cmd_add_attribute('Note', 'oldstyledefaultdate')
            mh.cmd_add_attribute('Note', 'newstyledefaultdate')
            self.assertIn('mydate', self.schema)
            self.assertIn('oldstyledefaultdate', self.schema)
            self.assertIn('newstyledefaultdate', self.schema)
            self.assertEqual(self.schema['mydate'].subjects(), ('Note', ))
            self.assertEqual(self.schema['mydate'].objects(), ('Date', ))
            testdate = date(2005, 12, 13)
            eid1 = mh.rqlexec('INSERT Note N')[0][0]
            eid2 = mh.rqlexec('INSERT Note N: N mydate %(mydate)s', {'mydate' : testdate})[0][0]
            d1 = mh.rqlexec('Any D WHERE X eid %(x)s, X mydate D', {'x': eid1})[0][0]
            d2 = mh.rqlexec('Any D WHERE X eid %(x)s, X mydate D', {'x': eid2})[0][0]
            d3 = mh.rqlexec('Any D WHERE X eid %(x)s, X oldstyledefaultdate D', {'x': eid1})[0][0]
            d4 = mh.rqlexec('Any D WHERE X eid %(x)s, X newstyledefaultdate D', {'x': eid1})[0][0]
            self.assertEqual(d1, date.today())
            self.assertEqual(d2, testdate)
            myfavoritedate = date(2013, 1, 1)
            self.assertEqual(d3, myfavoritedate)
            self.assertEqual(d4, myfavoritedate)

    def test_drop_chosen_constraints_ctxmanager(self):
        with self.mh() as (cnx, mh):
            with mh.cmd_dropped_constraints('Note', 'unique_id', UniqueConstraint):
                mh.cmd_add_attribute('Note', 'unique_id')
                # make sure the maxsize constraint is not dropped
                self.assertRaises(ValidationError,
                                  mh.rqlexec,
                                  'INSERT Note N: N unique_id "xyz"')
                mh.rollback()
                # make sure the unique constraint is dropped
                mh.rqlexec('INSERT Note N: N unique_id "x"')
                mh.rqlexec('INSERT Note N: N unique_id "x"')
                mh.rqlexec('DELETE Note N')

    def test_drop_required_ctxmanager(self):
        with self.mh() as (cnx, mh):
            with mh.cmd_dropped_constraints('Note', 'unique_id', cstrtype=None,
                                            droprequired=True):
                mh.cmd_add_attribute('Note', 'unique_id')
                mh.rqlexec('INSERT Note N')
                mh.rqlexec('SET N unique_id "x"')
            # make sure the required=True was restored
            self.assertRaises(ValidationError, mh.rqlexec, 'INSERT Note N')
            mh.rollback()

    def test_rename_attribute(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn('civility', self.schema)
            eid1 = mh.rqlexec('INSERT Personne X: X nom "lui", X sexe "M"')[0][0]
            eid2 = mh.rqlexec('INSERT Personne X: X nom "l\'autre", X sexe NULL')[0][0]
            mh.cmd_rename_attribute('Personne', 'sexe', 'civility')
            self.assertNotIn('sexe', self.schema)
            self.assertIn('civility', self.schema)
            # test data has been backported
            c1 = mh.rqlexec('Any C WHERE X eid %s, X civility C' % eid1)[0][0]
            self.assertEqual(c1, 'M')
            c2 = mh.rqlexec('Any C WHERE X eid %s, X civility C' % eid2)[0][0]
            self.assertEqual(c2, None)

    def test_workflow_actions(self):
        with self.mh() as (cnx, mh):
            wf = mh.cmd_add_workflow(u'foo', ('Personne', 'Email'),
                                     ensure_workflowable=False)
            for etype in ('Personne', 'Email'):
                s1 = mh.rqlexec('Any N WHERE WF workflow_of ET, ET name "%s", WF name N' %
                                etype)[0][0]
                self.assertEqual(s1, "foo")
                s1 = mh.rqlexec('Any N WHERE ET default_workflow WF, ET name "%s", WF name N' %
                                etype)[0][0]
                self.assertEqual(s1, "foo")

    def test_add_entity_type(self):
        with self.mh() as (cnx, mh):
            self.assertNotIn('Folder2', self.schema)
            self.assertNotIn('filed_under2', self.schema)
            mh.cmd_add_entity_type('Folder2')
            self.assertIn('Folder2', self.schema)
            self.assertIn('Old', self.schema)
            self.assertTrue(cnx.execute('CWEType X WHERE X name "Folder2"'))
            self.assertIn('filed_under2', self.schema)
            self.assertTrue(cnx.execute('CWRType X WHERE X name "filed_under2"'))
            self.assertEqual(sorted(str(rs) for rs in self.schema['Folder2'].subject_relations()),
                              ['created_by', 'creation_date', 'cw_source', 'cwuri',
                               'description', 'description_format',
                               'eid',
                               'filed_under2', 'has_text',
                               'identity', 'in_basket', 'inlined_rel', 'is', 'is_instance_of',
                               'modification_date', 'name', 'owned_by'])
            self.assertCountEqual([str(rs) for rs in self.schema['Folder2'].object_relations()],
                                  ['filed_under2', 'identity', 'inlined_rel'])
            # Old will be missing as it has been renamed into 'New' in the migrated
            # schema while New hasn't been added here.
            self.assertEqual(sorted(str(e) for e in self.schema['filed_under2'].subjects()),
                             sorted(str(e) for e in self.schema.entities() if not e.final and e != 'Old'))
            self.assertEqual(self.schema['filed_under2'].objects(), ('Folder2',))
            eschema = self.schema.eschema('Folder2')
            for cstr in eschema.rdef('name').constraints:
                self.assertTrue(hasattr(cstr, 'eid'))

    def test_add_entity_type_with_constraint(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type('Activity')
            constraints = self.table_constraints(mh, 'cw_Activity')
            self.assertEqual(len(constraints), 2, constraints)

    def test_add_cube_with_custom_final_type(self):
        with self.mh() as (cnx, mh):
            try:
                mh.cmd_add_cube('fakecustomtype')
                self.assertIn('Numeric', self.schema)
                self.assertTrue(self.schema['Numeric'].final)
                rdef = self.schema['num'].rdefs[('Location', 'Numeric')]
                self.assertEqual(rdef.scale, 10)
                self.assertEqual(rdef.precision, 18)
                fields = self.table_schema(mh, '%sLocation' % SQL_PREFIX)
                self.assertEqual(fields['%snum' % SQL_PREFIX], ('numeric', None)) # XXX
            finally:
                mh.cmd_drop_cube('fakecustomtype')
                mh.drop_entity_type('Numeric')

    def test_add_drop_entity_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type('Folder2')
            wf = mh.cmd_add_workflow(u'folder2 wf', 'Folder2',
                                     ensure_workflowable=False)
            todo = wf.add_state(u'todo', initial=True)
            done = wf.add_state(u'done')
            wf.add_transition(u'redoit', done, todo)
            wf.add_transition(u'markasdone', todo, done)
            cnx.commit()
            eschema = self.schema.eschema('Folder2')
            mh.cmd_drop_entity_type('Folder2')
            self.assertNotIn('Folder2', self.schema)
            self.assertFalse(cnx.execute('CWEType X WHERE X name "Folder2"'))
            # test automatic workflow deletion
            self.assertFalse(cnx.execute('Workflow X WHERE NOT X workflow_of ET'))
            self.assertFalse(cnx.execute('State X WHERE NOT X state_of WF'))
            self.assertFalse(cnx.execute('Transition X WHERE NOT X transition_of WF'))

    def test_rename_entity_type(self):
        with self.mh() as (cnx, mh):
            entity = mh.create_entity('Old', name=u'old')
            self.repo.type_from_eid(entity.eid, entity._cw)
            mh.cmd_rename_entity_type('Old', 'New')
            dbh = self.repo.system_source.dbhelper
            indices = set(dbh.list_indices(cnx.cnxset.cu, 'cw_New'))
            self.assertNotIn(build_index_name('cw_Old', ['cw_name']), indices)
            self.assertIn(build_index_name('cw_New', ['cw_name']), indices)
            mh.cmd_rename_attribute('New', 'name', 'new_name')
            indices = set(dbh.list_indices(cnx.cnxset.cu, 'cw_New'))
            # check 'indexed' index
            self.assertNotIn(build_index_name('cw_New', ['cw_name'], 'idx_'), indices)
            self.assertIn(build_index_name('cw_New', ['cw_new_name'], 'idx_'), indices)
            # check 'unique' index
            self.assertNotIn(build_index_name('cw_New', ['cw_name'], 'key_'), indices)
            self.assertIn(build_index_name('cw_New', ['cw_new_name'], 'key_'), indices)

    def test_add_drop_relation_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type('Folder2', auto=False)
            mh.cmd_add_relation_type('filed_under2')
            self.assertIn('filed_under2', self.schema)
            # Old will be missing as it has been renamed into 'New' in the migrated
            # schema while New hasn't been added here.
            self.assertEqual(sorted(str(e) for e in self.schema['filed_under2'].subjects()),
                             sorted(str(e) for e in self.schema.entities()
                                    if not e.final and e != 'Old'))
            self.assertEqual(self.schema['filed_under2'].objects(), ('Folder2',))
            mh.cmd_drop_relation_type('filed_under2')
            self.assertNotIn('filed_under2', self.schema)
            # this should not crash
            mh.cmd_drop_relation_type('filed_under2')

    def test_add_relation_definition_nortype(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_relation_definition('Personne', 'concerne2', 'Affaire')
            self.assertEqual(self.schema['concerne2'].subjects(),
                             ('Personne',))
            self.assertEqual(self.schema['concerne2'].objects(),
                             ('Affaire', ))
            self.assertEqual(self.schema['concerne2'].rdef('Personne', 'Affaire').cardinality,
                              '1*')
            mh.cmd_add_relation_definition('Personne', 'concerne2', 'Note')
            self.assertEqual(sorted(self.schema['concerne2'].objects()), ['Affaire', 'Note'])
            mh.create_entity('Personne', nom=u'tot')
            mh.create_entity('Affaire')
            mh.rqlexec('SET X concerne2 Y WHERE X is Personne, Y is Affaire')
            cnx.commit()
            mh.cmd_drop_relation_definition('Personne', 'concerne2', 'Affaire')
            self.assertIn('concerne2', self.schema)
            mh.cmd_drop_relation_definition('Personne', 'concerne2', 'Note')
            self.assertNotIn('concerne2', self.schema)

    def test_drop_relation_definition_existant_rtype(self):
        with self.mh() as (cnx, mh):
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].subjects()),
                              ['Affaire', 'Personne'])
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].objects()),
                              ['Affaire', 'Division', 'Note', 'Societe', 'SubDivision'])
            mh.cmd_drop_relation_definition('Personne', 'concerne', 'Affaire')
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].subjects()),
                              ['Affaire'])
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].objects()),
                              ['Division', 'Note', 'Societe', 'SubDivision'])
            mh.cmd_add_relation_definition('Personne', 'concerne', 'Affaire')
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].subjects()),
                              ['Affaire', 'Personne'])
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].objects()),
                              ['Affaire', 'Division', 'Note', 'Societe', 'SubDivision'])
            # trick: overwrite self.maxeid to avoid deletion of just reintroduced types
            self.maxeid = cnx.execute('Any MAX(X)')[0][0]

    def test_drop_relation_definition_with_specialization(self):
        with self.mh() as (cnx, mh):
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].subjects()),
                              ['Affaire', 'Personne'])
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].objects()),
                              ['Affaire', 'Division', 'Note', 'Societe', 'SubDivision'])
            mh.cmd_drop_relation_definition('Affaire', 'concerne', 'Societe')
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].subjects()),
                              ['Affaire', 'Personne'])
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].objects()),
                              ['Affaire', 'Note'])
            mh.cmd_add_relation_definition('Affaire', 'concerne', 'Societe')
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].subjects()),
                              ['Affaire', 'Personne'])
            self.assertEqual(sorted(str(e) for e in self.schema['concerne'].objects()),
                              ['Affaire', 'Division', 'Note', 'Societe', 'SubDivision'])
            # trick: overwrite self.maxeid to avoid deletion of just reintroduced types
            self.maxeid = cnx.execute('Any MAX(X)')[0][0]

    def test_rename_relation(self):
        self.skipTest('implement me')

    def test_change_relation_props_non_final(self):
        with self.mh() as (cnx, mh):
            rschema = self.schema['concerne']
            card = rschema.rdef('Affaire', 'Societe').cardinality
            self.assertEqual(card, '**')
            try:
                mh.cmd_change_relation_props('Affaire', 'concerne', 'Societe',
                                             cardinality='?*')
                card = rschema.rdef('Affaire', 'Societe').cardinality
                self.assertEqual(card, '?*')
            finally:
                mh.cmd_change_relation_props('Affaire', 'concerne', 'Societe',
                                             cardinality='**')

    def test_change_relation_props_final(self):
        with self.mh() as (cnx, mh):
            rschema = self.schema['adel']
            card = rschema.rdef('Personne', 'String').fulltextindexed
            self.assertEqual(card, False)
            try:
                mh.cmd_change_relation_props('Personne', 'adel', 'String',
                                             fulltextindexed=True)
                card = rschema.rdef('Personne', 'String').fulltextindexed
                self.assertEqual(card, True)
            finally:
                mh.cmd_change_relation_props('Personne', 'adel', 'String',
                                             fulltextindexed=False)

    def test_sync_schema_props_perms_rqlconstraints(self):
        with self.mh() as (cnx, mh):
            # Drop one of the RQLConstraint.
            rdef = self.schema['evaluee'].rdefs[('Personne', 'Note')]
            oldconstraints = rdef.constraints
            self.assertIn('S created_by U',
                          [cstr.expression for cstr in oldconstraints])
            mh.cmd_sync_schema_props_perms('evaluee', commit=True)
            newconstraints = rdef.constraints
            self.assertNotIn('S created_by U',
                             [cstr.expression for cstr in newconstraints])

            # Drop all RQLConstraint.
            rdef = self.schema['travaille'].rdefs[('Personne', 'Societe')]
            oldconstraints = rdef.constraints
            self.assertEqual(len(oldconstraints), 2)
            mh.cmd_sync_schema_props_perms('travaille', commit=True)
            rdef = self.schema['travaille'].rdefs[('Personne', 'Societe')]
            newconstraints = rdef.constraints
            self.assertEqual(len(newconstraints), 0)

    @tag('longrun')
    def test_sync_schema_props_perms(self):
        with self.mh() as (cnx, mh):
            nbrqlexpr_start = cnx.execute('Any COUNT(X) WHERE X is RQLExpression')[0][0]
            migrschema['titre'].rdefs[('Personne', 'String')].order = 7
            migrschema['adel'].rdefs[('Personne', 'String')].order = 6
            migrschema['ass'].rdefs[('Personne', 'String')].order = 5
            migrschema['Personne'].description = 'blabla bla'
            migrschema['titre'].description = 'usually a title'
            migrschema['titre'].rdefs[('Personne', 'String')].description = 'title for this person'
            delete_concerne_rqlexpr = self._rrqlexpr_rset(cnx, 'delete', 'concerne')
            add_concerne_rqlexpr = self._rrqlexpr_rset(cnx, 'add', 'concerne')

            # make sure properties (e.g. etype descriptions) are synced by the
            # second call to sync_schema
            mh.cmd_sync_schema_props_perms(syncprops=False, commit=False)
            mh.cmd_sync_schema_props_perms(commit=False)

            self.assertEqual(cnx.execute('Any D WHERE X name "Personne", X description D')[0][0],
                             'blabla bla')
            self.assertEqual(cnx.execute('Any D WHERE X name "titre", X description D')[0][0],
                             'usually a title')
            self.assertEqual(cnx.execute('Any D WHERE X relation_type RT, RT name "titre",'
                                         'X from_entity FE, FE name "Personne",'
                                         'X description D')[0][0],
                             'title for this person')
            rinorder = [n for n, in cnx.execute(
                'Any N ORDERBY O,N WHERE X is CWAttribute, X relation_type RT, RT name N,'
                'X from_entity FE, FE name "Personne",'
                'X ordernum O')]
            expected = [u'nom', u'prenom', u'sexe', u'promo', u'ass', u'adel', u'titre',
                        u'web', u'tel', u'fax', u'datenaiss', u'test', u'tzdatenaiss',
                        u'description', u'firstname', u'photo',
                        u'creation_date', u'cwuri', u'modification_date']
            self.assertEqual(expected, rinorder)

            # test permissions synchronization ####################################
            # new rql expr to add note entity
            eexpr = self._erqlexpr_entity(cnx, 'add', 'Note')
            self.assertEqual(eexpr.expression,
                              'X ecrit_part PE, U in_group G, '
                              'PE require_permission P, P name "add_note", P require_group G')
            self.assertEqual([et.name for et in eexpr.reverse_add_permission], ['Note'])
            self.assertEqual(eexpr.reverse_read_permission, ())
            self.assertEqual(eexpr.reverse_delete_permission, ())
            self.assertEqual(eexpr.reverse_update_permission, ())
            self.assertTrue(self._rrqlexpr_rset(cnx, 'add', 'para'))
            # no rqlexpr to delete para attribute
            self.assertFalse(self._rrqlexpr_rset(cnx, 'delete', 'para'))
            # new rql expr to add ecrit_par relation
            rexpr = self._rrqlexpr_entity(cnx, 'add', 'ecrit_par')
            self.assertEqual(rexpr.expression,
                              'O require_permission P, P name "add_note", '
                              'U in_group G, P require_group G')
            self.assertEqual([rdef.rtype.name for rdef in rexpr.reverse_add_permission], ['ecrit_par'])
            self.assertEqual(rexpr.reverse_read_permission, ())
            self.assertEqual(rexpr.reverse_delete_permission, ())
            # no more rqlexpr to delete and add travaille relation
            self.assertFalse(self._rrqlexpr_rset(cnx, 'add', 'travaille'))
            self.assertFalse(self._rrqlexpr_rset(cnx, 'delete', 'travaille'))
            # no more rqlexpr to delete and update Societe entity
            self.assertFalse(self._erqlexpr_rset(cnx, 'update', 'Societe'))
            self.assertFalse(self._erqlexpr_rset(cnx, 'delete', 'Societe'))
            # no more rqlexpr to read Affaire entity
            self.assertFalse(self._erqlexpr_rset(cnx, 'read', 'Affaire'))
            # rqlexpr to update Affaire entity has been updated
            eexpr = self._erqlexpr_entity(cnx, 'update', 'Affaire')
            self.assertEqual(eexpr.expression, 'X concerne S, S owned_by U')
            # no change for rqlexpr to add and delete Affaire entity
            self.assertEqual(len(self._erqlexpr_rset(cnx, 'delete', 'Affaire')), 1)
            self.assertEqual(len(self._erqlexpr_rset(cnx, 'add', 'Affaire')), 1)
            # no change for rqlexpr to add and delete concerne relation
            self.assertEqual(len(self._rrqlexpr_rset(cnx, 'delete', 'concerne')),
                             len(delete_concerne_rqlexpr))
            self.assertEqual(len(self._rrqlexpr_rset(cnx, 'add', 'concerne')),
                             len(add_concerne_rqlexpr))
            # * migrschema involve:
            #   * 7 erqlexprs deletions (2 in (Affaire + Societe + Note.para) + 1 Note.something
            #   * 2 rrqlexprs deletions (travaille)
            #   * 1 update (Affaire update)
            #   * 2 new (Note add, ecrit_par add)
            #   * 2 implicit new for attributes (Note.para, Person.test)
            # remaining orphan rql expr which should be deleted at commit (composite relation)
            # unattached expressions -> pending deletion on commit
            self.assertEqual(cnx.execute('Any COUNT(X) WHERE X is RQLExpression, X exprtype "ERQLExpression",'
                                         'NOT ET1 read_permission X, NOT ET2 add_permission X, '
                                         'NOT ET3 delete_permission X, NOT ET4 update_permission X')[0][0],
                              7)
            self.assertEqual(cnx.execute('Any COUNT(X) WHERE X is RQLExpression, X exprtype "RRQLExpression",'
                                         'NOT ET1 read_permission X, NOT ET2 add_permission X, '
                                         'NOT ET3 delete_permission X, NOT ET4 update_permission X')[0][0],
                              2)
            # finally
            self.assertEqual(cnx.execute('Any COUNT(X) WHERE X is RQLExpression')[0][0],
                             nbrqlexpr_start + 1 + 2 + 2 + 2)
            cnx.commit()
            # unique_together test
            self.assertEqual(len(self.schema.eschema('Personne')._unique_together), 1)
            self.assertCountEqual(self.schema.eschema('Personne')._unique_together[0],
                                               ('nom', 'prenom', 'datenaiss'))
            rset = cnx.execute('Any C WHERE C is CWUniqueTogetherConstraint, C constraint_of ET, ET name "Personne"')
            self.assertEqual(len(rset), 1)
            relations = [r.name for r in rset.get_entity(0, 0).relations]
            self.assertCountEqual(relations, ('nom', 'prenom', 'datenaiss'))

            # serialized constraint changed
            constraints = self.table_constraints(mh, 'cw_Personne')
            self.assertEqual(len(constraints), 1, constraints)
            rdef = migrschema['promo'].rdefs['Personne', 'String']
            cstr = rdef.constraint_by_type('StaticVocabularyConstraint')
            self.assertIn(cstr.name_for(rdef), constraints)

    def _erqlexpr_rset(self, cnx, action, ertype):
        rql = 'RQLExpression X WHERE ET is CWEType, ET %s_permission X, ET name %%(name)s' % action
        return cnx.execute(rql, {'name': ertype})

    def _erqlexpr_entity(self, cnx, action, ertype):
        rset = self._erqlexpr_rset(cnx, action, ertype)
        self.assertEqual(len(rset), 1)
        return rset.get_entity(0, 0)

    def _rrqlexpr_rset(self, cnx, action, ertype):
        rql = 'RQLExpression X WHERE RT is CWRType, RDEF %s_permission X, RT name %%(name)s, RDEF relation_type RT' % action
        return cnx.execute(rql, {'name': ertype})

    def _rrqlexpr_entity(self, cnx, action, ertype):
        rset = self._rrqlexpr_rset(cnx, action, ertype)
        self.assertEqual(len(rset), 1)
        return rset.get_entity(0, 0)

    def test_set_size_constraint(self):
        with self.mh() as (cnx, mh):
            # existing previous value
            try:
                mh.cmd_set_size_constraint('CWEType', 'name', 128)
            finally:
                mh.cmd_set_size_constraint('CWEType', 'name', 64)
            # non existing previous value
            try:
                mh.cmd_set_size_constraint('CWEType', 'description', 256)
            finally:
                mh.cmd_set_size_constraint('CWEType', 'description', None)

    @tag('longrun')
    def test_add_drop_cube_and_deps(self):
        with self.mh() as (cnx, mh):
            schema = self.repo.schema
            self.assertEqual(sorted((str(s), str(o)) for s, o in schema['see_also'].rdefs),
                             sorted([('EmailThread', 'EmailThread'), ('Folder', 'Folder'),
                                     ('Bookmark', 'Bookmark'), ('Bookmark', 'Note'),
                                     ('Note', 'Note'), ('Note', 'Bookmark')]))
            try:
                mh.cmd_drop_cube('fakeemail', removedeps=True)
                # file was there because it's an email dependancy, should have been removed
                self.assertNotIn('fakeemail', self.config.cubes())
                self.assertNotIn(self.config.cube_dir('fakeemail'), self.config.cubes_path())
                self.assertNotIn('file', self.config.cubes())
                self.assertNotIn(self.config.cube_dir('file'), self.config.cubes_path())
                for ertype in ('Email', 'EmailThread', 'EmailPart', 'File',
                               'sender', 'in_thread', 'reply_to', 'data_format'):
                    self.assertNotIn(ertype, schema)
                self.assertEqual(sorted(schema['see_also'].rdefs),
                                  sorted([('Folder', 'Folder'),
                                          ('Bookmark', 'Bookmark'),
                                          ('Bookmark', 'Note'),
                                          ('Note', 'Note'),
                                          ('Note', 'Bookmark')]))
                self.assertEqual(sorted(schema['see_also'].subjects()), ['Bookmark', 'Folder', 'Note'])
                self.assertEqual(sorted(schema['see_also'].objects()), ['Bookmark', 'Folder', 'Note'])
                self.assertEqual(cnx.execute('Any X WHERE X pkey "system.version.fakeemail"').rowcount, 0)
                self.assertEqual(cnx.execute('Any X WHERE X pkey "system.version.file"').rowcount, 0)
            finally:
                mh.cmd_add_cube('fakeemail')
                self.assertIn('fakeemail', self.config.cubes())
                self.assertIn(self.config.cube_dir('fakeemail'), self.config.cubes_path())
                self.assertIn('file', self.config.cubes())
                self.assertIn(self.config.cube_dir('file'), self.config.cubes_path())
                for ertype in ('Email', 'EmailThread', 'EmailPart', 'File',
                               'sender', 'in_thread', 'reply_to', 'data_format'):
                    self.assertIn(ertype, schema)
                self.assertEqual(sorted(schema['see_also'].rdefs),
                                 sorted([('EmailThread', 'EmailThread'), ('Folder', 'Folder'),
                                         ('Bookmark', 'Bookmark'),
                                         ('Bookmark', 'Note'),
                                         ('Note', 'Note'),
                                         ('Note', 'Bookmark')]))
                self.assertEqual(sorted(schema['see_also'].subjects()), ['Bookmark', 'EmailThread', 'Folder', 'Note'])
                self.assertEqual(sorted(schema['see_also'].objects()), ['Bookmark', 'EmailThread', 'Folder', 'Note'])
                from cubes.fakeemail.__pkginfo__ import version as email_version
                from cubes.file.__pkginfo__ import version as file_version
                self.assertEqual(cnx.execute('Any V WHERE X value V, X pkey "system.version.fakeemail"')[0][0],
                                  email_version)
                self.assertEqual(cnx.execute('Any V WHERE X value V, X pkey "system.version.file"')[0][0],
                                  file_version)
                # trick: overwrite self.maxeid to avoid deletion of just reintroduced
                #        types (and their associated tables!)
                self.maxeid = cnx.execute('Any MAX(X)')[0][0]
                # why this commit is necessary is unclear to me (though without it
                # next test may fail complaining of missing tables
                cnx.commit()


    @tag('longrun')
    def test_add_drop_cube_no_deps(self):
        with self.mh() as (cnx, mh):
            cubes = set(self.config.cubes())
            schema = self.repo.schema
            try:
                mh.cmd_drop_cube('fakeemail')
                cubes.remove('fakeemail')
                self.assertNotIn('fakeemail', self.config.cubes())
                self.assertIn('file', self.config.cubes())
                for ertype in ('Email', 'EmailThread', 'EmailPart',
                               'sender', 'in_thread', 'reply_to'):
                    self.assertNotIn(ertype, schema)
            finally:
                mh.cmd_add_cube('fakeemail')
                self.assertIn('fakeemail', self.config.cubes())
                # trick: overwrite self.maxeid to avoid deletion of just reintroduced
                #        types (and their associated tables!)
                self.maxeid = cnx.execute('Any MAX(X)')[0][0] # XXXXXXX KILL KENNY
                # why this commit is necessary is unclear to me (though without it
                # next test may fail complaining of missing tables
                cnx.commit()

    def test_drop_dep_cube(self):
        with self.mh() as (cnx, mh):
            with self.assertRaises(ConfigurationError) as cm:
                mh.cmd_drop_cube('file')
            self.assertEqual(str(cm.exception), "can't remove cube file, used as a dependency")

    @tag('longrun')
    def test_introduce_base_class(self):
        with self.mh() as (cnx, mh):
            mh.cmd_add_entity_type('Para')
            self.assertEqual(sorted(et.type for et in self.schema['Para'].specialized_by()),
                             ['Note'])
            self.assertEqual(self.schema['Note'].specializes().type, 'Para')
            mh.cmd_add_entity_type('Text')
            self.assertEqual(sorted(et.type for et in self.schema['Para'].specialized_by()),
                              ['Note', 'Text'])
            self.assertEqual(self.schema['Text'].specializes().type, 'Para')
            # test columns have been actually added
            text = cnx.execute('INSERT Text X: X para "hip", X summary "hop", X newattr "momo"').get_entity(0, 0)
            note = cnx.execute('INSERT Note X: X para "hip", X shortpara "hop", X newattr "momo", X unique_id "x"').get_entity(0, 0)
            aff = cnx.execute('INSERT Affaire X').get_entity(0, 0)
            self.assertTrue(cnx.execute('SET X newnotinlined Y WHERE X eid %(x)s, Y eid %(y)s',
                                         {'x': text.eid, 'y': aff.eid}))
            self.assertTrue(cnx.execute('SET X newnotinlined Y WHERE X eid %(x)s, Y eid %(y)s',
                                         {'x': note.eid, 'y': aff.eid}))
            self.assertTrue(cnx.execute('SET X newinlined Y WHERE X eid %(x)s, Y eid %(y)s',
                                         {'x': text.eid, 'y': aff.eid}))
            self.assertTrue(cnx.execute('SET X newinlined Y WHERE X eid %(x)s, Y eid %(y)s',
                                         {'x': note.eid, 'y': aff.eid}))
            # XXX remove specializes by ourselves, else tearDown fails when removing
            # Para because of Note inheritance. This could be fixed by putting the
            # MemSchemaCWETypeDel(session, name) operation in the
            # after_delete_entity(CWEType) hook, since in that case the MemSchemaSpecializesDel
            # operation would be removed before, but I'm not sure this is a desired behaviour.
            #
            # also we need more tests about introducing/removing base classes or
            # specialization relationship...
            cnx.execute('DELETE X specializes Y WHERE Y name "Para"')
            cnx.commit()
            self.assertEqual(sorted(et.type for et in self.schema['Para'].specialized_by()),
                             [])
            self.assertEqual(self.schema['Note'].specializes(), None)
            self.assertEqual(self.schema['Text'].specializes(), None)

    def test_add_symmetric_relation_type(self):
        with self.mh() as (cnx, mh):
            self.assertFalse(self.table_sql(mh, 'same_as_relation'))
            mh.cmd_add_relation_type('same_as')
            self.assertTrue(self.table_sql(mh, 'same_as_relation'))

    def test_change_attribute_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_create_entity('Societe', tel=1)
            mh.commit()
            mh.change_attribute_type('Societe', 'tel', 'Float')
            self.assertNotIn(('Societe', 'Int'), self.schema['tel'].rdefs)
            self.assertIn(('Societe', 'Float'), self.schema['tel'].rdefs)
            self.assertEqual(self.schema['tel'].rdefs[('Societe', 'Float')].object, 'Float')
            tel = mh.rqlexec('Any T WHERE X tel T')[0][0]
            self.assertEqual(tel, 1.0)
            self.assertIsInstance(tel, float)

    def test_drop_required_inlined_relation(self):
        with self.mh() as (cnx, mh):
            bob = mh.cmd_create_entity('Personne', nom=u'bob')
            note = mh.cmd_create_entity('Note', ecrit_par=bob)
            mh.commit()
            rdef = mh.fs_schema.rschema('ecrit_par').rdefs[('Note', 'Personne')]
            with tempattr(rdef, 'cardinality', '1*'):
                mh.sync_schema_props_perms('ecrit_par', syncperms=False)
            mh.cmd_drop_relation_type('ecrit_par')
            self.assertNotIn('%secrit_par' % SQL_PREFIX,
                             self.table_schema(mh, '%sPersonne' % SQL_PREFIX))

    def test_drop_inlined_rdef_delete_data(self):
        with self.mh() as (cnx, mh):
            note = mh.cmd_create_entity('Note', ecrit_par=cnx.user.eid)
            mh.commit()
            mh.drop_relation_definition('Note', 'ecrit_par', 'CWUser')
            self.assertFalse(mh.sqlexec('SELECT * FROM cw_Note WHERE cw_ecrit_par IS NOT NULL'))

    def test_storage_changed(self):
        with self.mh() as (cnx, mh):
            john = mh.cmd_create_entity('Personne', nom=u'john',
                                        photo=Binary(b'something'))
            bill = mh.cmd_create_entity('Personne', nom=u'bill')
            mh.commit()
            with TemporaryDirectory() as tempdir:
                bfs_storage = storages.BytesFileSystemStorage(tempdir)
                storages.set_attribute_storage(self.repo, 'Personne', 'photo', bfs_storage)
                mh.cmd_storage_changed('Personne', 'photo')
                bob = mh.cmd_create_entity('Personne', nom=u'bob')
                bffss_dir_content = os.listdir(tempdir)
                self.assertEqual(len(bffss_dir_content), 1)
                john.cw_clear_all_caches()
                self.assertEqual(john.photo.getvalue(),
                                 osp.join(tempdir,
                                          bffss_dir_content[0]).encode('utf8'))
                bob.cw_clear_all_caches()
                self.assertIsNone(bob.photo)
                bill.cw_clear_all_caches()
                self.assertIsNone(bill.photo)
                storages.unset_attribute_storage(self.repo, 'Personne', 'photo')


class MigrationCommandsComputedTC(MigrationTC):
    """ Unit tests for computed relations and attributes
    """
    appid = 'datacomputed'

    def setUp(self):
        MigrationTC.setUp(self)
        # ensure vregistry is reloaded, needed by generated hooks for computed
        # attributes
        self.repo.vreg.set_schema(self.repo.schema)

    def test_computed_relation_add_relation_definition(self):
        self.assertNotIn('works_for', self.schema)
        with self.mh() as (cnx, mh):
            with self.assertRaises(ExecutionError) as exc:
                mh.cmd_add_relation_definition('Employee', 'works_for', 'Company')
        self.assertEqual(str(exc.exception),
                         'Cannot add a relation definition for a computed '
                         'relation (works_for)')

    def test_computed_relation_drop_relation_definition(self):
        self.assertIn('notes', self.schema)
        with self.mh() as (cnx, mh):
            with self.assertRaises(ExecutionError) as exc:
                mh.cmd_drop_relation_definition('Company', 'notes', 'Note')
        self.assertEqual(str(exc.exception),
                         'Cannot drop a relation definition for a computed '
                         'relation (notes)')

    def test_computed_relation_add_relation_type(self):
        self.assertNotIn('works_for', self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_add_relation_type('works_for')
            self.assertIn('works_for', self.schema)
            self.assertEqual(self.schema['works_for'].rule,
                             'O employees S, NOT EXISTS (O associates S)')
            self.assertEqual(self.schema['works_for'].objects(), ('Company',))
            self.assertEqual(self.schema['works_for'].subjects(), ('Employee',))
            self.assertFalse(self.table_sql(mh, 'works_for_relation'))
            e = cnx.create_entity('Employee')
            a = cnx.create_entity('Employee')
            cnx.create_entity('Company', employees=e, associates=a)
            cnx.commit()
            company = cnx.execute('Company X').get_entity(0, 0)
            self.assertEqual([e.eid],
                             [x.eid for x in company.reverse_works_for])
            mh.rollback()

    def test_computed_relation_drop_relation_type(self):
        self.assertIn('notes', self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_relation_type('notes')
        self.assertNotIn('notes', self.schema)

    def test_computed_relation_sync_schema_props_perms(self):
        self.assertIn('whatever', self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_sync_schema_props_perms('whatever')
            self.assertEqual(self.schema['whatever'].rule,
                             'S employees E, O associates E')
            self.assertEqual(self.schema['whatever'].objects(), ('Company',))
            self.assertEqual(self.schema['whatever'].subjects(), ('Company',))
            self.assertFalse(self.table_sql(mh, 'whatever_relation'))

    def test_computed_relation_sync_schema_props_perms_security(self):
        with self.mh() as (cnx, mh):
            rdef = next(iter(self.schema['perm_changes'].rdefs.values()))
            self.assertEqual(rdef.permissions,
                             {'add': (), 'delete': (),
                              'read': ('managers', 'users')})
            mh.cmd_sync_schema_props_perms('perm_changes')
            self.assertEqual(self.schema['perm_changes'].permissions,
                             {'read': ('managers',)})
            rdef = next(iter(self.schema['perm_changes'].rdefs.values()))
            self.assertEqual(rdef.permissions,
                             {'add': (), 'delete': (),
                              'read': ('managers',)})

    def test_computed_relation_sync_schema_props_perms_on_rdef(self):
        self.assertIn('whatever', self.schema)
        with self.mh() as (cnx, mh):
            with self.assertRaises(ExecutionError) as exc:
                mh.cmd_sync_schema_props_perms(
                    ('Company', 'whatever', 'Person'))
        self.assertEqual(str(exc.exception),
                         'Cannot synchronize a relation definition for a computed '
                         'relation (whatever)')

    def test_computed_relation_rename_relation_type(self):
        with self.mh() as (cnx, mh):
            mh.cmd_rename_relation_type('to_be_renamed', 'renamed')
        self.assertIn('renamed', self.schema)
        self.assertNotIn('to_be_renamed', self.schema)

    # computed attributes migration ############################################

    def setup_add_score(self):
        with self.admin_access.client_cnx() as cnx:
            assert not cnx.execute('Company X')
            c = cnx.create_entity('Company')
            e1 = cnx.create_entity('Employee', reverse_employees=c)
            cnx.create_entity('Note', note=2, concerns=e1)
            e2 = cnx.create_entity('Employee', reverse_employees=c)
            cnx.create_entity('Note', note=4, concerns=e2)
            cnx.commit()

    def assert_score_initialized(self, mh):
        self.assertEqual(self.schema['score'].rdefs['Company', 'Float'].formula,
                         'Any AVG(NN) WHERE X employees E, N concerns E, N note NN')
        fields = self.table_schema(mh, '%sCompany' % SQL_PREFIX)
        self.assertEqual(fields['%sscore' % SQL_PREFIX], ('double precision', None))
        self.assertEqual([[3.0]],
                         mh.rqlexec('Any CS WHERE C score CS, C is Company').rows)

    def test_computed_attribute_add_relation_type(self):
        self.assertNotIn('score', self.schema)
        self.setup_add_score()
        with self.mh() as (cnx, mh):
            mh.cmd_add_relation_type('score')
            self.assertIn('score', self.schema)
            self.assertEqual(self.schema['score'].objects(), ('Float',))
            self.assertEqual(self.schema['score'].subjects(), ('Company',))
            self.assert_score_initialized(mh)

    def test_computed_attribute_add_attribute(self):
        self.assertNotIn('score', self.schema)
        self.setup_add_score()
        with self.mh() as (cnx, mh):
            mh.cmd_add_attribute('Company', 'score')
            self.assertIn('score', self.schema)
            self.assert_score_initialized(mh)

    def assert_computed_attribute_dropped(self):
        self.assertNotIn('note20', self.schema)
        with self.mh() as (cnx, mh):
            fields = self.table_schema(mh, '%sNote' % SQL_PREFIX)
        self.assertNotIn('%snote20' % SQL_PREFIX, fields)

    def test_computed_attribute_drop_type(self):
        self.assertIn('note20', self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_relation_type('note20')
        self.assert_computed_attribute_dropped()

    def test_computed_attribute_drop_relation_definition(self):
        self.assertIn('note20', self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_relation_definition('Note', 'note20', 'Int')
        self.assert_computed_attribute_dropped()

    def test_computed_attribute_drop_attribute(self):
        self.assertIn('note20', self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_drop_attribute('Note', 'note20')
        self.assert_computed_attribute_dropped()

    def test_computed_attribute_sync_schema_props_perms_rtype(self):
        self.assertIn('note100', self.schema)
        with self.mh() as (cnx, mh):
            mh.cmd_sync_schema_props_perms('note100')
        rdef = self.schema['note100'].rdefs['Note', 'Int']
        self.assertEqual(rdef.formula_select.as_string(),
                         'Any (N * 100) WHERE X note N, X is Note')
        self.assertEqual(rdef.formula, 'Any N*100 WHERE X note N')

    def test_computed_attribute_sync_schema_props_perms_rdef(self):
        self.setup_add_score()
        with self.mh() as (cnx, mh):
            mh.cmd_sync_schema_props_perms(('Note', 'note100', 'Int'))
            self.assertEqual([[200], [400]],
                             cnx.execute('Any N ORDERBY N WHERE X note100 N').rows)
            self.assertEqual([[300]],
                             cnx.execute('Any CS WHERE C score100 CS, C is Company').rows)


if __name__ == '__main__':
    unittest_main()



from unittest import SkipTest

from cubicweb.devtools import qunit

from os import path as osp


class JScript(qunit.QUnitTestCase):

    timeout_error = SkipTest
    all_js_tests = (
        ("/static/jstests/test_utils.js", (
            "/data/cubicweb.js",
            "/data/cubicweb.compat.js",
            "/data/cubicweb.python.js",
            "/static/jstests/utils.js",
            ),
         ),

        ("/static/jstests/test_htmlhelpers.js", (
            "/data/cubicweb.js",
            "/data/cubicweb.compat.js",
            "/data/cubicweb.python.js",
            "/data/cubicweb.htmlhelpers.js",
            ),
         ),

        ("/static/jstests/test_ajax.js", (
            "/data/cubicweb.python.js",
            "/data/cubicweb.js",
            "/data/cubicweb.compat.js",
            "/data/cubicweb.htmlhelpers.js",
            "/data/cubicweb.ajax.js",
            ),
         ),
    )


if __name__ == '__main__':
    from unittest import main
    main()



# -*- coding: utf-8 -*-
# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""
.. _ResourceMode:

Resource mode
-------------

Standard resource mode
``````````````````````

A resource *mode* is a predefined set of settings for various resources
directories, such as cubes, instances, etc. to ease development with the
framework. There are two running modes with *CubicWeb*:

* **system**: resources are searched / created in the system directories (eg
  usually requiring root access):

  - instances are stored in :file:`<INSTALL_PREFIX>/etc/cubicweb.d`
  - temporary files (such as pid file) in :file:`<INSTALL_PREFIX>/var/run/cubicweb`

  where `<INSTALL_PREFIX>` is the detected installation prefix ('/usr/local' for
  instance).

* **user**: resources are searched / created in the user home directory:

  - instances are stored in :file:`~/etc/cubicweb.d`
  - temporary files (such as pid file) in :file:`/tmp`


.. _CubicwebWithinVirtualEnv:

Within virtual environment
``````````````````````````

If you are not administrator of you machine or if you need to play with some
specific version of |cubicweb| you can use virtualenv_ a tool to create
isolated Python environments.

- instances are stored in :file:`<VIRTUAL_ENV>/etc/cubicweb.d`
- temporary files (such as pid file) in :file:`<VIRTUAL_ENV>/var/run/cubicweb`

.. _virtualenv: http://pypi.python.org/pypi/virtualenv


Custom resource location
````````````````````````

Notice that each resource path may be explicitly set using an environment
variable if the default doesn't suit your needs. Here are the default resource
directories that are affected according to mode:

* **system**: ::

        CW_INSTANCES_DIR = <INSTALL_PREFIX>/etc/cubicweb.d/
        CW_INSTANCES_DATA_DIR = <INSTALL_PREFIX>/var/lib/cubicweb/instances/
        CW_RUNTIME_DIR = <INSTALL_PREFIX>/var/run/cubicweb/

* **user**: ::

        CW_INSTANCES_DIR = ~/etc/cubicweb.d/
        CW_INSTANCES_DATA_DIR = ~/etc/cubicweb.d/
        CW_RUNTIME_DIR = /tmp

Cubes search path is also affected, see the :ref:`Cube` section.


Setting Cubicweb Mode
`````````````````````

By default, the mode is set to 'system' for standard installation. The mode is
set to 'user' if `cubicweb is used from a mercurial repository`_. You can force
this by setting the :envvar:`CW_MODE` environment variable to either 'user' or
'system' so you can easily:

* use system wide installation but user specific instances and all, without root
  privileges on the system (`export CW_MODE=user`)

* use local checkout of cubicweb on system wide instances (requires root
  privileges on the system (`export CW_MODE=system`)

If you've a doubt about the mode you're currently running, check the first line
outputed by the :command:`cubicweb-ctl list` command.

.. _`cubicweb is used from a mercurial repository`: CubicwebDevelopmentMod_


.. _CubicwebDevelopmentMod:

Development Mode (source)
`````````````````````````

If :file:`.hg` directory is found into the cubicweb package, there are
specific resource rules.

`<CW_SOFTWARE_ROOT>` is the source checkout's ``cubicweb`` directory:

* main cubes directory is `<CW_SOFTWARE_ROOT>/../../cubes`. You can specify
  another one with :envvar:`CW_INSTANCES_DIR` environment variable or simply
  add some other directories by using :envvar:`CW_CUBES_PATH`

* cubicweb migration files are searched in `<CW_SOFTWARE_ROOT>/misc/migration`
  instead of `<INSTALL_PREFIX>/share/cubicweb/migration/`.


Development Mode (virtualenv)
`````````````````````````````

If a virtualenv is found to be activated (i.e. a VIRTUAL_ENV variable is found
in environment), the virtualenv root is used as `<INSTALL_PREFIX>`. This, in
particular, makes it possible to work in `setuptools development mode`_
(``python setup.py develop``) without any further configuration.

.. _`setuptools development mode`: https://pythonhosted.org/setuptools/setuptools.html#development-mode

.. _ConfigurationEnv:

Environment configuration
-------------------------

Python
``````

If you installed *CubicWeb* by cloning the Mercurial shell repository or from source
distribution, then you will need to update the environment variable PYTHONPATH by
adding the path to `cubicweb`:

Add the following lines to either :file:`.bashrc` or :file:`.bash_profile` to
configure your development environment ::

    export PYTHONPATH=/full/path/to/grshell-cubicweb

If you installed *CubicWeb* with packages, no configuration is required and your
new cubes will be placed in `/usr/share/cubicweb/cubes` and your instances will
be placed in `/etc/cubicweb.d`.


CubicWeb
````````

Here are all environment variables that may be used to configure *CubicWeb*:

.. envvar:: CW_MODE

   Resource mode: user or system, as explained in :ref:`ResourceMode`.

.. envvar:: CW_CUBES_PATH

   Augments the default search path for cubes. You may specify several
   directories using ':' as separator (';' under windows environment).

.. envvar:: CW_INSTANCES_DIR

   Directory where cubicweb instances will be found.

.. envvar:: CW_INSTANCES_DATA_DIR

   Directory where cubicweb instances data will be written (backup file...)

.. envvar:: CW_RUNTIME_DIR

   Directory where pid files will be written
"""

from __future__ import print_function

import importlib
import logging
import logging.config
import os
from os.path import (exists, join, expanduser, abspath, normpath,
                     basename, isdir, dirname, splitext, realpath)
import pkgutil
import pkg_resources
import re
from smtplib import SMTP
import stat
import sys
from threading import Lock
from warnings import filterwarnings

from six import text_type

from logilab.common.decorators import cached, classproperty
from logilab.common.deprecation import deprecated
from logilab.common.logging_ext import set_log_methods, init_log
from logilab.common.configuration import (Configuration, Method,
                                          ConfigurationMixIn, merge_options,
                                          _validate as lgc_validate)

from cubicweb import (CW_SOFTWARE_ROOT, CW_MIGRATION_MAP,
                      ConfigurationError, Binary, _)
from cubicweb.toolsutils import create_dir, option_value_from_env

CONFIGURATIONS = []

SMTP_LOCK = Lock()


def configuration_cls(name):
    """return the configuration class registered with the given name"""
    try:
        return [c for c in CONFIGURATIONS if c.name == name][0]
    except IndexError:
        raise ConfigurationError('no such config %r (check it exists with "cubicweb-ctl list")' % name)


def possible_configurations(directory):
    """return a list of installed configurations in a directory
    according to \*-ctl files
    """
    return [name for name in ('repository', 'all-in-one', 'pyramid')
            if exists(join(directory, '%s.conf' % name))]


def guess_configuration(directory):
    """try to guess the configuration to use for a directory. If multiple
    configurations are found, ConfigurationError is raised
    """
    modes = possible_configurations(directory)
    if len(modes) != 1:
        raise ConfigurationError('unable to guess configuration from %r %s'
                                 % (directory, modes))
    return modes[0]


def _find_prefix(start_path=None):
    """Return the prefix path of CubicWeb installation.

    Walk parent directories of `start_path` looking for one containing a
    'share/cubicweb' directory. The first matching directory is assumed as the
    prefix installation of CubicWeb.

    If run from within a virtualenv, the virtualenv root is used as
    `start_path`. Otherwise, `start_path` defaults to cubicweb package
    directory path.
    """
    if start_path is None:
        try:
            prefix = os.environ['VIRTUAL_ENV']
        except KeyError:
            prefix = CW_SOFTWARE_ROOT
    else:
        prefix = start_path
    if not isdir(prefix):
        prefix = dirname(prefix)
    old_prefix = None
    while (not isdir(join(prefix, 'share', 'cubicweb'))
           or prefix.endswith('.egg')):
        if prefix == old_prefix:
            return sys.prefix
        old_prefix = prefix
        prefix = dirname(prefix)
    return prefix


def _cube_pkgname(cube):
    if not cube.startswith('cubicweb_'):
        return 'cubicweb_' + cube
    return cube


def _expand_modname(modname):
    """expand modules names `modname` if exists by walking non package submodules
    and yield (submodname, filepath) including `modname` itself

    If the file ends with .pyc or .pyo (python bytecode) also check that the
    corresponding source .py file exists before yielding.
    """
    try:
        loader = pkgutil.find_loader(modname)
    except ImportError:
        return
    if not loader:
        return

    def check_source_file(filepath):
        if filepath[-4:] in ('.pyc', '.pyo'):
            if not exists(filepath[:-1]):
                return False
        return True

    filepath = loader.get_filename()
    if not check_source_file(filepath):
        return
    yield modname, filepath
    if loader.is_package(modname):
        path = dirname(filepath)
        for subloader, subname, ispkg in pkgutil.walk_packages([path]):
            # ignore subpackages (historical behavior)
            if not ispkg:
                filepath = subloader.find_module(subname).get_filename()
                if check_source_file(filepath):
                    yield modname + '.' + subname, filepath


# persistent options definition
PERSISTENT_OPTIONS = (
    ('encoding',
     {'type' : 'string',
      'default': 'UTF-8',
      'help': _('user interface encoding'),
      'group': 'ui', 'sitewide': True,
      }),
    ('language',
     {'type' : 'string',
      'default': 'en',
      'vocabulary': Method('available_languages'),
      'help': _('language of the user interface'),
      'group': 'ui',
      }),
    ('date-format',
     {'type' : 'string',
      'default': '%Y/%m/%d',
      'help': _('how to format date in the ui (see <a href="http://docs.python.org/library/datetime.html#strftime-strptime-behavior">this page</a> for format description)'),
      'group': 'ui',
      }),
    ('datetime-format',
     {'type' : 'string',
      'default': '%Y/%m/%d %H:%M',
      'help': _('how to format date and time in the ui (see <a href="http://docs.python.org/library/datetime.html#strftime-strptime-behavior">this page</a> for format description)'),
      'group': 'ui',
      }),
    ('time-format',
     {'type' : 'string',
      'default': '%H:%M',
      'help': _('how to format time in the ui (see <a href="http://docs.python.org/library/datetime.html#strftime-strptime-behavior">this page</a> for format description)'),
      'group': 'ui',
      }),
    ('float-format',
     {'type' : 'string',
      'default': '%.3f',
      'help': _('how to format float numbers in the ui'),
      'group': 'ui',
      }),
    ('default-text-format',
     {'type' : 'choice',
      'choices': ('text/plain', 'text/rest', 'text/html', 'text/markdown'),
      'default': 'text/plain',
      'help': _('default text format for rich text fields.'),
      'group': 'ui',
      }),
    ('short-line-size',
     {'type' : 'int',
      'default': 80,
      'help': _('maximum number of characters in short description'),
      'group': 'navigation',
      }),
    )

def register_persistent_options(options):
    global PERSISTENT_OPTIONS
    PERSISTENT_OPTIONS = merge_options(PERSISTENT_OPTIONS + options)

CFGTYPE2ETYPE_MAP = {
    'string': 'String',
    'choice': 'String',
    'yn':     'Boolean',
    'int':    'Int',
    'float' : 'Float',
    }

_forced_mode = os.environ.get('CW_MODE')
assert _forced_mode in (None, 'system', 'user')

# CWDEV tells whether directories such as i18n/, web/data/, etc. (ie containing
# some other resources than python libraries) are located with the python code
# or as a 'shared' cube
CWDEV = exists(join(CW_SOFTWARE_ROOT, 'i18n'))

try:
    _INSTALL_PREFIX = os.environ['CW_INSTALL_PREFIX']
except KeyError:
    _INSTALL_PREFIX = _find_prefix()
_USR_INSTALL = _INSTALL_PREFIX == '/usr'

class CubicWebNoAppConfiguration(ConfigurationMixIn):
    """base class for cubicweb configuration without a specific instance directory
    """
    # to set in concrete configuration
    name = None
    # log messages format (see logging module documentation for available keys)
    log_format = '%(asctime)s - (%(name)s) %(levelname)s: %(message)s'
    # the format below can be useful to debug multi thread issues:
    # log_format = '%(asctime)s - [%(threadName)s] (%(name)s) %(levelname)s: %(message)s'
    # nor remove appobjects based on unused interface [???]
    cleanup_unused_appobjects = True

    quick_start = False

    if 'VIRTUAL_ENV' in os.environ:
        mode = _forced_mode or 'user'
        _CUBES_DIR = join(_INSTALL_PREFIX, 'share', 'cubicweb', 'cubes')
    elif CWDEV and _forced_mode != 'system':
        mode = 'user'
        _CUBES_DIR = join(CW_SOFTWARE_ROOT, '../../cubes')
    else:
        mode = _forced_mode or 'system'
        _CUBES_DIR = join(_INSTALL_PREFIX, 'share', 'cubicweb', 'cubes')

    CUBES_DIR = realpath(abspath(os.environ.get('CW_CUBES_DIR', _CUBES_DIR)))
    CUBES_PATH = os.environ.get('CW_CUBES_PATH', '').split(os.pathsep)

    options = (
       ('log-threshold',
         {'type' : 'string', # XXX use a dedicated type?
          'default': 'WARNING',
          'help': 'server\'s log level',
          'group': 'main', 'level': 1,
          }),
        ('umask',
         {'type' : 'int',
          'default': 0o077,
          'help': 'permission umask for files created by the server',
          'group': 'main', 'level': 2,
          }),
        # common configuration options which are potentially required as soon as
        # you're using "base" application objects (ie to really server/web
        # specific)
        ('base-url',
         {'type' : 'string',
          'default': None,
          'help': 'web server root url',
          'group': 'main', 'level': 1,
          }),
        ('allow-email-login',
         {'type' : 'yn',
          'default': False,
          'help': 'allow users to login with their primary email if set',
          'group': 'main', 'level': 2,
          }),
        ('mangle-emails',
         {'type' : 'yn',
          'default': False,
          'help': "don't display actual email addresses but mangle them if \
this option is set to yes",
          'group': 'email', 'level': 3,
          }),
        )

    def __getitem__(self, key):
        """Get configuration option, by first looking at environmnent."""
        file_value = super(CubicWebNoAppConfiguration, self).__getitem__(key)
        value = option_value_from_env(key, file_value)
        if value is not None:
            option_def = self.get_option_def(key)
            value = lgc_validate(value, option_def)
        return value

    # static and class methods used to get instance independant resources ##
    @staticmethod
    def cubicweb_version():
        """return installed cubicweb version"""
        from logilab.common.changelog import Version
        from cubicweb import __pkginfo__
        version = __pkginfo__.numversion
        assert len(version) == 3, version
        return Version(version)

    @staticmethod
    def persistent_options_configuration():
        return Configuration(options=PERSISTENT_OPTIONS)

    @classmethod
    def shared_dir(cls):
        """return the shared data directory (i.e. directory where standard
        library views and data may be found)
        """
        if CWDEV:
            return join(CW_SOFTWARE_ROOT, 'web')
        return cls.cube_dir('shared')

    @classmethod
    def i18n_lib_dir(cls):
        """return instance's i18n directory"""
        if CWDEV:
            return join(CW_SOFTWARE_ROOT, 'i18n')
        return join(cls.shared_dir(), 'i18n')

    @classmethod
    def cw_languages(cls):
        for fname in os.listdir(join(cls.i18n_lib_dir())):
            if fname.endswith('.po'):
                yield splitext(fname)[0]


    @classmethod
    def available_cubes(cls):
        cubes = set()
        prefix = 'cubicweb_'
        for entry_point in pkg_resources.iter_entry_points(
                group='cubicweb.cubes', name=None):
            try:
                module = entry_point.load()
            except ImportError:
                continue
            else:
                modname = module.__name__
                if not modname.startswith(prefix):
                    cls.warning('entry point %s does not appear to be a cube',
                                entry_point)
                    continue
                cubes.add(modname[len(prefix):])
        # Legacy cubes.
        for directory in cls.cubes_search_path():
            if not exists(directory):
                cls.error('unexistant directory in cubes search path: %s'
                          % directory)
                continue
            for cube in os.listdir(directory):
                if cube == 'shared':
                    continue
                if not re.match('[_A-Za-z][_A-Za-z0-9]*$', cube):
                    continue # skip invalid python package name
                if cube == 'pyramid':
                    cls._warn_pyramid_cube()
                    continue
                cubedir = join(directory, cube)
                if isdir(cubedir) and exists(join(cubedir, '__init__.py')):
                    cubes.add(cube)

        def sortkey(cube):
            """Preserve sorting with "cubicweb_" prefix."""
            prefix = 'cubicweb_'
            if cube.startswith(prefix):
                # add a suffix to have a deterministic sorting between
                # 'cubicweb_<cube>' and '<cube>' (useful in tests with "hash
                # randomization" turned on).
                return cube[len(prefix):] + '~'
            return cube

        return sorted(cubes, key=sortkey)

    @classmethod
    def cubes_search_path(cls):
        """return the path of directories where cubes should be searched"""
        path = [realpath(abspath(normpath(directory))) for directory in cls.CUBES_PATH
                if directory.strip() and exists(directory.strip())]
        if not cls.CUBES_DIR in path and exists(cls.CUBES_DIR):
            path.append(cls.CUBES_DIR)
        return path

    @classproperty
    def extrapath(cls):
        extrapath = {}
        for cubesdir in cls.cubes_search_path():
            if cubesdir != cls.CUBES_DIR:
                extrapath[cubesdir] = 'cubes'
        return extrapath

    @classmethod
    def cube_dir(cls, cube):
        """return the cube directory for the given cube id, raise
        `ConfigurationError` if it doesn't exist
        """
        pkgname = _cube_pkgname(cube)
        loader = pkgutil.find_loader(pkgname)
        if loader:
            return dirname(loader.get_filename())
        # Legacy cubes.
        for directory in cls.cubes_search_path():
            cubedir = join(directory, cube)
            if exists(cubedir):
                return cubedir
        msg = 'no module %(pkg)s in search path nor cube %(cube)r in %(path)s'
        raise ConfigurationError(msg % {'cube': cube,
                                        'pkg': _cube_pkgname(cube),
                                        'path': cls.cubes_search_path()})

    @classmethod
    def cube_migration_scripts_dir(cls, cube):
        """cube migration scripts directory"""
        return join(cls.cube_dir(cube), 'migration')

    @classmethod
    def cube_pkginfo(cls, cube):
        """return the information module for the given cube"""
        pkgname = _cube_pkgname(cube)
        try:
            return importlib.import_module('%s.__pkginfo__' % pkgname)
        except ImportError:
            cube = CW_MIGRATION_MAP.get(cube, cube)
            try:
                parent = __import__('cubes.%s.__pkginfo__' % cube)
                return getattr(parent, cube).__pkginfo__
            except Exception as ex:
                raise ConfigurationError(
                    'unable to find packaging information for cube %s (%s: %s)'
                    % (cube, ex.__class__.__name__, ex))

    @classmethod
    def cube_version(cls, cube):
        """return the version of the cube located in the given directory
        """
        from logilab.common.changelog import Version
        version = cls.cube_pkginfo(cube).numversion
        assert len(version) == 3, version
        return Version(version)

    @classmethod
    def _cube_deps(cls, cube, key, oldkey):
        """return cubicweb cubes used by the given cube"""
        pkginfo = cls.cube_pkginfo(cube)
        try:
            # explicit __xxx_cubes__ attribute
            deps = getattr(pkginfo, key)
        except AttributeError:
            # deduce cubes from generic __xxx__ attribute
            try:
                gendeps = getattr(pkginfo, key.replace('_cubes', ''))
            except AttributeError:
                deps = {}
            else:
                deps = dict( (x[len('cubicweb-'):], v)
                             for x, v in gendeps.items()
                             if x.startswith('cubicweb-'))
        for depcube in deps:
            try:
                newname = CW_MIGRATION_MAP[depcube]
            except KeyError:
                pass
            else:
                deps[newname] = deps.pop(depcube)
        return deps

    @classmethod
    def cube_depends_cubicweb_version(cls, cube):
        # XXX no backward compat (see _cube_deps above)
        try:
            pkginfo = cls.cube_pkginfo(cube)
            deps = getattr(pkginfo, '__depends__')
            return deps.get('cubicweb')
        except AttributeError:
            return None

    @classmethod
    def cube_dependencies(cls, cube):
        """return cubicweb cubes used by the given cube"""
        return cls._cube_deps(cube, '__depends_cubes__', '__use__')

    @classmethod
    def cube_recommends(cls, cube):
        """return cubicweb cubes recommended by the given cube"""
        return cls._cube_deps(cube, '__recommends_cubes__', '__recommend__')

    @classmethod
    def expand_cubes(cls, cubes, with_recommends=False):
        """expand the given list of top level cubes used by adding recursivly
        each cube dependencies
        """
        cubes = list(cubes)
        todo = cubes[:]
        if with_recommends:
            available = set(cls.available_cubes())
        while todo:
            cube = todo.pop(0)
            for depcube in cls.cube_dependencies(cube):
                if depcube not in cubes:
                    cubes.append(depcube)
                    todo.append(depcube)
            if with_recommends:
                for depcube in cls.cube_recommends(cube):
                    if depcube not in cubes and depcube in available:
                        cubes.append(depcube)
                        todo.append(depcube)
        return cubes

    @classmethod
    def reorder_cubes(cls, cubes):
        """reorder cubes from the top level cubes to inner dependencies
        cubes
        """
        from logilab.common.graph import ordered_nodes, UnorderableGraph
        graph = {}
        for cube in cubes:
            cube = CW_MIGRATION_MAP.get(cube, cube)
            graph[cube] = set(dep for dep in cls.cube_dependencies(cube)
                              if dep in cubes)
            graph[cube] |= set(dep for dep in cls.cube_recommends(cube)
                               if dep in cubes)
        try:
            return ordered_nodes(graph)
        except UnorderableGraph as ex:
            raise ConfigurationError(ex)

    @classmethod
    def cls_adjust_sys_path(cls):
        """update python path if necessary"""
        from cubicweb import _CubesImporter
        _CubesImporter.install()
        import cubes
        cubes.__path__ = cls.cubes_search_path()

    @classmethod
    def load_available_configs(cls):
        for confmod in ('web.webconfig',  'etwist.twconfig',
                        'server.serverconfig', 'pyramid.config'):
            try:
                __import__('cubicweb.%s' % confmod)
            except ImportError as exc:
                cls.warning('failed to load config module %s (%s)',
                            confmod, exc)

    @classmethod
    def load_cwctl_plugins(cls):
        cls.cls_adjust_sys_path()
        for ctlmod in ('web.webctl',  'etwist.twctl', 'server.serverctl',
                       'devtools.devctl', 'pyramid.pyramidctl'):
            try:
                __import__('cubicweb.%s' % ctlmod)
            except ImportError as exc:
                cls.warning('failed to load cubicweb-ctl plugin %s (%s)',
                            ctlmod, exc)
                continue
            cls.info('loaded cubicweb-ctl plugin %s', ctlmod)
        for cube in cls.available_cubes():
            cubedir = cls.cube_dir(cube)
            pluginfile = join(cubedir, 'ccplugin.py')
            initfile = join(cubedir, '__init__.py')
            if cube.startswith('cubicweb_'):
                pkgname = cube
            else:
                pkgname = 'cubes.%s' % cube
            if exists(pluginfile):
                try:
                    __import__(pkgname + '.ccplugin')
                    cls.info('loaded cubicweb-ctl plugin from %s', cube)
                except Exception:
                    cls.exception('while loading plugin %s', pluginfile)
            elif exists(initfile):
                try:
                    __import__(pkgname)
                except Exception:
                    cls.exception('while loading cube %s', cube)
            else:
                cls.warning('no __init__ file in cube %s', cube)

    @classmethod
    def init_available_cubes(cls):
        """cubes may register some sources (svnfile for instance) in their
        __init__ file, so they should be loaded early in the startup process
        """
        for cube in cls.available_cubes():
            try:
                __import__('cubes.%s' % cube)
            except Exception as ex:
                cls.warning("can't init cube %s: %s", cube, ex)

    cubicweb_appobject_path = set(['entities'])
    cube_appobject_path = set(['entities'])

    def __init__(self, debugmode=False):
        if debugmode:
            # in python 2.7, DeprecationWarning are not shown anymore by default
            filterwarnings('default', category=DeprecationWarning)
        register_stored_procedures()
        self._cubes = None
        super(CubicWebNoAppConfiguration, self).__init__()
        self.debugmode = debugmode
        self.adjust_sys_path()
        self.load_defaults()
        # will be properly initialized later by _gettext_init
        self.translations = {'en': (text_type, lambda ctx, msgid: text_type(msgid) )}
        self._site_loaded = set()
        # don't register ReStructured Text directives by simple import, avoid pb
        # with eg sphinx.
        # XXX should be done properly with a function from cw.uicfg
        try:
            from cubicweb.ext.rest import cw_rest_init
        except ImportError:
            pass
        else:
            cw_rest_init()

    def adjust_sys_path(self):
        # overriden in CubicWebConfiguration
        self.cls_adjust_sys_path()

    def init_log(self, logthreshold=None, logfile=None, syslog=False):
        """init the log service"""
        if logthreshold is None:
            if self.debugmode:
                logthreshold = 'DEBUG'
            else:
                logthreshold = self['log-threshold']
        if sys.platform == 'win32':
            # no logrotate on win32, so use logging rotation facilities
            # for now, hard code weekly rotation every sunday, and 52 weeks kept
            # idea: make this configurable?
            init_log(self.debugmode, syslog, logthreshold, logfile, self.log_format,
                     rotation_parameters={'when': 'W6', # every sunday
                                          'interval': 1,
                                          'backupCount': 52})
        else:
            init_log(self.debugmode, syslog, logthreshold, logfile, self.log_format)
        # configure simpleTal logger
        logging.getLogger('simpleTAL').setLevel(logging.ERROR)

    def schema_modnames(self):
        modnames = []
        for name in ('bootstrap', 'base', 'workflow', 'Bookmark'):
            modnames.append(('cubicweb', 'cubicweb.schemas.' + name))
        for cube in reversed(self.cubes()):
            for modname, filepath in _expand_modname('cubes.{0}.schema'.format(cube)):
                modnames.append((cube, modname))
        if self.apphome:
            apphome = realpath(self.apphome)
            for modname, filepath in _expand_modname('schema'):
                if realpath(filepath).startswith(apphome):
                    modnames.append(('data', modname))
        return modnames

    def appobjects_modnames(self):
        """return a list of modules where the registry will look for
        application objects. By default return nothing in NoApp config.
        """
        return []

    apphome = None

    def load_site_cubicweb(self, cubes=()):
        """load site_cubicweb file for `cubes`"""
        for cube in reversed(cubes or self.cubes()):
            if cube in self._site_loaded:
                continue
            try:
                self._load_site_cubicweb(cube)
                self._site_loaded.add(cube)
            except ImportError:
                continue
        if self.apphome is not None:
            # Would occur, e.g., upon `cubicweb-ctl i18ncube <cube>`.
            self._load_site_cubicweb(None)

    def _load_site_cubicweb(self, cube):
        """Load site_cubicweb.py from `cube` (or apphome if cube is None)."""
        if cube is not None:
            try:
                modname = 'cubicweb_%s' % cube
                __import__(modname)
            except ImportError:
                modname = 'cubes.%s' % cube
                __import__(modname)
            modname = modname + '.site_cubicweb'
            __import__(modname)
            return sys.modules[modname]
        else:
            import imp
            apphome_site = join(self.apphome, 'site_cubicweb.py')
            if exists(apphome_site):
                with open(apphome_site, 'rb') as f:
                    return imp.load_source('site_cubicweb', apphome_site, f)

    def cwproperty_definitions(self):
        cfg = self.persistent_options_configuration()
        for section, options in cfg.options_by_section():
            section = section.lower()
            for optname, optdict, value in options:
                key = '%s.%s' % (section, optname)
                type, vocab = self.map_option(optdict)
                default = cfg.option_default(optname, optdict)
                pdef = {'type': type, 'vocabulary': vocab, 'default': default,
                        'help': optdict['help'],
                        'sitewide': optdict.get('sitewide', False)}
                yield key, pdef

    def map_option(self, optdict):
        try:
            vocab = optdict['choices']
        except KeyError:
            vocab = optdict.get('vocabulary')
            if isinstance(vocab, Method):
                vocab = getattr(self, vocab.method, ())
        return CFGTYPE2ETYPE_MAP[optdict['type']], vocab

    def default_instance_id(self):
        """return the instance identifier, useful for option which need this
        as default value
        """
        return None

    _cubes = None

    @classmethod
    def _warn_pyramid_cube(cls):
        cls.warning("cubicweb-pyramid got integrated into CubicWeb; "
                    "remove it from your project's dependencies")

    def init_cubes(self, cubes):
        cubes = list(cubes)
        if 'pyramid' in cubes:
            self._warn_pyramid_cube()
            cubes.remove('pyramid')
        self._cubes = self.reorder_cubes(cubes)
        # load cubes'__init__.py file first
        for cube in cubes:
            try:
                importlib.import_module(_cube_pkgname(cube))
            except ImportError:
                # Legacy cube.
                __import__('cubes.%s' % cube)
        self.load_site_cubicweb()

    def cubes(self):
        """return the list of cubes used by this instance

        result is ordered from the top level cubes to inner dependencies
        cubes
        """
        assert self._cubes is not None, 'cubes not initialized'
        return self._cubes

    def cubes_path(self):
        """return the list of path to cubes used by this instance, from outer
        most to inner most cubes
        """
        return [self.cube_dir(p) for p in self.cubes()]

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    @classmethod
    def debug(cls, msg, *a, **kw):
        pass
    info = warning = error = critical = exception = debug


class CubicWebConfiguration(CubicWebNoAppConfiguration):
    """base class for cubicweb server and web configurations"""

    if CubicWebNoAppConfiguration.mode == 'user':
        _INSTANCES_DIR = expanduser('~/etc/cubicweb.d/')
    #mode == system'
    elif _USR_INSTALL:
        _INSTANCES_DIR = '/etc/cubicweb.d/'
    else:
        _INSTANCES_DIR = join(_INSTALL_PREFIX, 'etc', 'cubicweb.d')

    # set to true during repair (shell, migration) to allow some things which
    # wouldn't be possible otherwise
    repairing = False

    # set by upgrade command
    verbosity = 0
    cmdline_options = None
    options = CubicWebNoAppConfiguration.options + (
        ('log-file',
         {'type' : 'string',
          'default': Method('default_log_file'),
          'help': 'file where output logs should be written',
          'group': 'main', 'level': 2,
          }),
        ('statsd-endpoint',
         {'type' : 'string',
          'default': '',
          'help': 'UDP address of the statsd endpoint; it must be formatted'
                  'like <ip>:<port>; disabled is unset.',
          'group': 'main', 'level': 2,
          }),
        # email configuration
        ('smtp-host',
         {'type' : 'string',
          'default': 'mail',
          'help': 'hostname of the SMTP mail server',
          'group': 'email', 'level': 1,
          }),
        ('smtp-port',
         {'type' : 'int',
          'default': 25,
          'help': 'listening port of the SMTP mail server',
          'group': 'email', 'level': 1,
          }),
        ('sender-name',
         {'type' : 'string',
          'default': Method('default_instance_id'),
          'help': 'name used as HELO name for outgoing emails from the \
repository.',
          'group': 'email', 'level': 2,
          }),
        ('sender-addr',
         {'type' : 'string',
          'default': 'cubicweb@mydomain.com',
          'help': 'email address used as HELO address for outgoing emails from \
the repository',
          'group': 'email', 'level': 1,
          }),
        ('logstat-interval',
         {'type' : 'int',
          'default': 0,
          'help': 'interval (in seconds) at which stats are dumped in the logstat file; set 0 to disable',
          'group': 'main', 'level': 2,
          }),
        ('logstat-file',
         {'type' : 'string',
          'default': Method('default_stats_file'),
          'help': 'file where stats for the instance should be written',
          'group': 'main', 'level': 2,
          }),
        )

    @classmethod
    def instances_dir(cls):
        """return the control directory"""
        return abspath(os.environ.get('CW_INSTANCES_DIR', cls._INSTANCES_DIR))

    @classmethod
    def migration_scripts_dir(cls):
        """cubicweb migration scripts directory"""
        if CWDEV:
            return join(CW_SOFTWARE_ROOT, 'misc', 'migration')
        mdir = join(_INSTALL_PREFIX, 'share', 'cubicweb', 'migration')
        if not exists(mdir):
            raise ConfigurationError('migration path %s doesn\'t exist' % mdir)
        return mdir

    @classmethod
    def config_for(cls, appid, config=None, debugmode=False, creating=False):
        """return a configuration instance for the given instance identifier
        """
        cls.load_available_configs()
        config = config or guess_configuration(cls.instance_home(appid))
        configcls = configuration_cls(config)
        return configcls(appid, debugmode, creating)

    @classmethod
    def possible_configurations(cls, appid):
        """return the name of possible configurations for the given
        instance id
        """
        home = cls.instance_home(appid)
        return possible_configurations(home)

    @classmethod
    def instance_home(cls, appid):
        """return the home directory of the instance with the given
        instance id
        """
        home = join(cls.instances_dir(), appid)
        if not exists(home):
            raise ConfigurationError('no such instance %s (check it exists with'
                                     ' "cubicweb-ctl list")' % appid)
        return home

    MODES = ('common', 'repository', 'Any')
    MCOMPAT = {'all-in-one': MODES,
               'repository': ('common', 'repository', 'Any')}
    @classmethod
    def accept_mode(cls, mode):
        #assert mode in cls.MODES, mode
        return mode in cls.MCOMPAT[cls.name]

    # default configuration methods ###########################################

    def default_instance_id(self):
        """return the instance identifier, useful for option which need this
        as default value
        """
        return self.appid

    def default_log_file(self):
        """return default path to the log file of the instance'server"""
        if self.mode == 'user':
            import tempfile
            basepath = join(tempfile.gettempdir(), '%s-%s' % (
                basename(self.appid), self.name))
            path = basepath + '.log'
            i = 1
            while exists(path) and i < 100: # arbitrary limit to avoid infinite loop
                try:
                    open(path, 'a')
                    break
                except IOError:
                    path = '%s-%s.log' % (basepath, i)
                    i += 1
            return path
        if _USR_INSTALL:
            return '/var/log/cubicweb/%s-%s.log' % (self.appid, self.name)
        else:
            log_path = os.path.join(_INSTALL_PREFIX, 'var', 'log', 'cubicweb', '%s-%s.log')
            return log_path % (self.appid, self.name)

    def default_stats_file(self):
        """return default path to the stats file of the instance'server"""
        logfile = self.default_log_file()
        if logfile.endswith('.log'):
            logfile = logfile[:-4]
        return logfile + '.stats'

    def default_pid_file(self):
        """return default path to the pid file of the instance'server"""
        if self.mode == 'system':
            if _USR_INSTALL:
                default = '/var/run/cubicweb/'
            else:
                default = os.path.join(_INSTALL_PREFIX, 'var', 'run', 'cubicweb')
        else:
            import tempfile
            default = tempfile.gettempdir()
        # runtime directory created on startup if necessary, don't check it
        # exists
        rtdir = abspath(os.environ.get('CW_RUNTIME_DIR', default))
        return join(rtdir, '%s-%s.pid' % (self.appid, self.name))

    # config -> repository

    def repository(self, vreg=None):
        """Return a new bootstrapped repository."""
        from cubicweb.server.repository import Repository
        repo = Repository(self, vreg=vreg)
        repo.bootstrap()
        return repo

    # instance methods used to get instance specific resources #############

    def __init__(self, appid, debugmode=False, creating=False):
        self.appid = appid
        # set to true while creating an instance
        self.creating = creating
        super(CubicWebConfiguration, self).__init__(debugmode)
        fake_gettext = (text_type, lambda ctx, msgid: text_type(msgid))
        for lang in self.available_languages():
            self.translations[lang] = fake_gettext
        self._cubes = None
        self.load_file_configuration(self.main_config_file())

    def adjust_sys_path(self):
        super(CubicWebConfiguration, self).adjust_sys_path()
        # adding apphome to python path is not usually necessary in production
        # environments, but necessary for tests
        if self.apphome and self.apphome not in sys.path:
            sys.path.insert(0, self.apphome)

    @property
    def apphome(self):
        return join(self.instances_dir(), self.appid)

    @property
    def appdatahome(self):
        if self.mode == 'system':
            if _USR_INSTALL:
                iddir = os.path.join('/var','lib', 'cubicweb', 'instances')
            else:
                iddir = os.path.join(_INSTALL_PREFIX, 'var', 'lib', 'cubicweb', 'instances')
        else:
            iddir = self.instances_dir()
        iddir = abspath(os.environ.get('CW_INSTANCES_DATA_DIR', iddir))
        return join(iddir, self.appid)

    def init_cubes(self, cubes):
        super(CubicWebConfiguration, self).init_cubes(cubes)
        # reload config file in cases options are defined in cubes __init__
        # or site_cubicweb files
        self.load_file_configuration(self.main_config_file())
        # configuration initialization hook
        self.load_configuration(**(self.cmdline_options or {}))

    def add_cubes(self, cubes):
        """add given cubes to the list of used cubes"""
        if not isinstance(cubes, list):
            cubes = list(cubes)
        self._cubes = self.reorder_cubes(list(self._cubes) + cubes)
        self.load_site_cubicweb(cubes)

    def main_config_file(self):
        """return instance's control configuration file"""
        return join(self.apphome, '%s.conf' % self.name)

    def save(self):
        """write down current configuration"""
        with open(self.main_config_file(), 'w') as fobj:
            self.generate_config(fobj)

    def check_writeable_uid_directory(self, path):
        """check given directory path exists, belongs to the user running the
        server process and is writeable.

        If not, try to fix this, letting exception propagate when not possible.
        """
        if not exists(path):
            self.info('creating %s directory', path)
            try:
                os.makedirs(path)
            except OSError as ex:
                self.warning('error while creating %s directory: %s', path, ex)
                return
        self.ensure_uid(path)

    def get_uid(self):
        if self['uid']:
            try:
                uid = int(self['uid'])
            except ValueError:
                from pwd import getpwnam
                uid = getpwnam(self['uid']).pw_uid
        else:
            try:
                uid = os.getuid()
            except AttributeError: # we are on windows
                return
        return uid

    def ensure_uid(self, path, enforce_write=False):
        if not exists(path):
            return
        uid = self.get_uid()
        if uid is None:
            return
        fstat = os.stat(path)
        if fstat.st_uid != uid:
            self.info('giving ownership of %s to %s', path, self['uid'])
            try:
                os.chown(path, uid, os.getgid())
            except OSError as ex:
                self.warning('error while giving ownership of %s to %s: %s',
                             path, self['uid'], ex)

        if enforce_write and not (fstat.st_mode & stat.S_IWUSR):
            self.info('forcing write permission on %s', path)
            try:
                os.chmod(path, fstat.st_mode | stat.S_IWUSR)
            except OSError as ex:
                self.warning('error while forcing write permission on %s: %s',
                             path, ex)

    def ensure_uid_directory(self, path, enforce_write=False):
        self.check_writeable_uid_directory(path)
        for dirpath, dirnames, filenames in os.walk(path):
            for name in filenames:
                self.ensure_uid(join(dirpath, name), enforce_write)
        return path

    @cached
    def instance_md5_version(self):
        from hashlib import md5 # pylint: disable=E0611
        infos = []
        for pkg in sorted(self.cubes()):
            version = self.cube_version(pkg)
            infos.append('%s-%s' % (pkg, version))
        infos.append('cubicweb-%s' % str(self.cubicweb_version()))
        return md5((';'.join(infos)).encode('ascii')).hexdigest()

    def load_configuration(self, **kw):
        """load instance's configuration files"""
        super(CubicWebConfiguration, self).load_configuration(**kw)
        if self.apphome and not self.creating:
            # init gettext
            self._gettext_init()

    def _load_site_cubicweb(self, cube):
        # overridden to register cube specific options
        mod = super(CubicWebConfiguration, self)._load_site_cubicweb(cube)
        if getattr(mod, 'options', None):
            self.register_options(mod.options)
            self.load_defaults()

    def init_log(self, logthreshold=None, force=False):
        """init the log service"""
        if not force and hasattr(self, '_logging_initialized'):
            return
        self._logging_initialized = True
        super_self = super(CubicWebConfiguration, self)
        super_self.init_log(logthreshold, logfile=self.get('log-file'))
        # read a config file if it exists
        logconfig = join(self.apphome, 'logging.conf')
        if exists(logconfig):
            logging.config.fileConfig(logconfig)
        # set the statsd address, if any
        if self.get('statsd-endpoint'):
            try:
                address, port = self.get('statsd-endpoint').split(':')
                port = int(port)
            except:
                self.error('statsd-endpoint: invalid address format ({}); '
                           'it should be "ip:port"'.format(self.get('statsd-endpoint')))
            else:
                import statsd_logger
                statsd_logger.setup('cubicweb.%s' % self.appid, (address, port))

    def available_languages(self, *args):
        """return available translation for an instance, by looking for
        compiled catalog

        take \*args to be usable as a vocabulary method
        """
        from glob import glob
        yield 'en' # ensure 'en' is yielded even if no .mo found
        for path in glob(join(self.apphome, 'i18n',
                              '*', 'LC_MESSAGES')):
            lang = path.split(os.sep)[-2]
            if lang != 'en':
                yield lang

    def _gettext_init(self):
        """set language for gettext"""
        from cubicweb.cwgettext import translation
        path = join(self.apphome, 'i18n')
        for language in self.available_languages():
            self.info("loading language %s", language)
            try:
                tr = translation('cubicweb', path, languages=[language])
                self.translations[language] = (tr.ugettext, tr.upgettext)
            except IOError:
                if self.mode != 'test':
                    # in test contexts, data/i18n does not exist, hence
                    # logging will only pollute the logs
                    self.exception('localisation support error for language %s',
                                   language)

    @staticmethod
    def _sorted_appobjects(appobjects):
        appobjects = sorted(appobjects)
        try:
            index = appobjects.index('entities')
        except ValueError:
            pass
        else:
            # put entities first
            appobjects.insert(0, appobjects.pop(index))
        return appobjects

    def appobjects_cube_modnames(self, cube):
        modnames = []
        cube_submodnames = self._sorted_appobjects(self.cube_appobject_path)
        for name in cube_submodnames:
            for modname, filepath in _expand_modname('.'.join(['cubes', cube, name])):
                modnames.append(modname)
        return modnames

    def appobjects_modnames(self):
        modnames = []
        for name in self._sorted_appobjects(self.cubicweb_appobject_path):
            for modname, filepath in _expand_modname('cubicweb.' + name):
                modnames.append(modname)
        for cube in reversed(self.cubes()):
            modnames.extend(self.appobjects_cube_modnames(cube))
        if self.apphome:
            cube_submodnames = self._sorted_appobjects(self.cube_appobject_path)
            apphome = realpath(self.apphome)
            for name in cube_submodnames:
                for modname, filepath in _expand_modname(name):
                    # ensure file is in apphome
                    if realpath(filepath).startswith(apphome):
                        modnames.append(modname)
        return modnames

    def set_sources_mode(self, sources):
        if not 'all' in sources:
            print('warning: ignoring specified sources, requires a repository '
                  'configuration')

    def i18ncompile(self, langs=None):
        from cubicweb import i18n
        if langs is None:
            langs = self.available_languages()
        i18ndir = join(self.apphome, 'i18n')
        if not exists(i18ndir):
            create_dir(i18ndir)
        sourcedirs = [join(path, 'i18n') for path in self.cubes_path()]
        sourcedirs.append(self.i18n_lib_dir())
        return i18n.compile_i18n_catalogs(sourcedirs, i18ndir, langs)

    def sendmails(self, msgs, fromaddr=None):
        """msgs: list of 2-uple (message object, recipients). Return False
        if connection to the smtp server failed, else True.
        """
        server, port = self['smtp-host'], self['smtp-port']
        if fromaddr is None:
            fromaddr = '%s <%s>' % (self['sender-name'], self['sender-addr'])
        SMTP_LOCK.acquire()
        try:
            try:
                smtp = SMTP(server, port)
            except Exception as ex:
                self.exception("can't connect to smtp server %s:%s (%s)",
                               server, port, ex)
                return False
            for msg, recipients in msgs:
                try:
                    smtp.sendmail(fromaddr, recipients, msg.as_string())
                except Exception as ex:
                    self.exception("error sending mail to %s (%s)",
                                   recipients, ex)
            smtp.close()
        finally:
            SMTP_LOCK.release()
        return True

set_log_methods(CubicWebNoAppConfiguration,
                logging.getLogger('cubicweb.configuration'))

# alias to get a configuration instance from an instance id
instance_configuration = CubicWebConfiguration.config_for
application_configuration = deprecated('use instance_configuration')(instance_configuration)


_EXT_REGISTERED = False
def register_stored_procedures():
    from logilab.database import FunctionDescr
    from rql.utils import register_function, iter_funcnode_variables
    from rql.nodes import SortTerm, Constant, VariableRef

    global _EXT_REGISTERED
    if _EXT_REGISTERED:
        return
    _EXT_REGISTERED = True

    class COMMA_JOIN(FunctionDescr):
        supported_backends = ('postgres', 'sqlite',)
        rtype = 'String'

        def st_description(self, funcnode, mainindex, tr):
            return ', '.join(sorted(term.get_description(mainindex, tr)
                                    for term in iter_funcnode_variables(funcnode)))

    register_function(COMMA_JOIN)  # XXX do not expose?


    class CONCAT_STRINGS(COMMA_JOIN):
        aggregat = True

    register_function(CONCAT_STRINGS) # XXX bw compat


    class GROUP_CONCAT(CONCAT_STRINGS):
        supported_backends = ('mysql', 'postgres', 'sqlite',)

    register_function(GROUP_CONCAT)


    class LIMIT_SIZE(FunctionDescr):
        supported_backends = ('postgres', 'sqlite',)
        minargs = maxargs = 3
        rtype = 'String'

        def st_description(self, funcnode, mainindex, tr):
            return funcnode.children[0].get_description(mainindex, tr)

    register_function(LIMIT_SIZE)


    class TEXT_LIMIT_SIZE(LIMIT_SIZE):
        supported_backends = ('mysql', 'postgres', 'sqlite',)
        minargs = maxargs = 2

    register_function(TEXT_LIMIT_SIZE)


    class FTIRANK(FunctionDescr):
        """return ranking of a variable that must be used as some has_text
        relation subject in the query's restriction. Usually used to sort result
        of full-text search by ranking.
        """
        supported_backends = ('postgres',)
        rtype = 'Float'

        def st_check_backend(self, backend, funcnode):
            """overriden so that on backend not supporting fti ranking, the
            function is removed when in an orderby clause, or replaced by a 1.0
            constant.
            """
            if not self.supports(backend):
                parent = funcnode.parent
                while parent is not None and not isinstance(parent, SortTerm):
                    parent = parent.parent
                if isinstance(parent, SortTerm):
                    parent.parent.remove(parent)
                else:
                    funcnode.parent.replace(funcnode, Constant(1.0, 'Float'))
                    parent = funcnode
                for vref in parent.iget_nodes(VariableRef):
                    vref.unregister_reference()

    register_function(FTIRANK)


    class FSPATH(FunctionDescr):
        """return path of some bytes attribute stored using the Bytes
        File-System Storage (bfss)
        """
        rtype = 'Bytes' # XXX return a String? potential pb with fs encoding

        def update_cb_stack(self, stack):
            assert len(stack) == 1
            stack[0] = self.source_execute

        def as_sql(self, backend, args):
            raise NotImplementedError(
                'This callback is only available for BytesFileSystemStorage '
                'managed attribute. Is FSPATH() argument BFSS managed?')

        def source_execute(self, source, session, value):
            fpath = source.binary_to_str(value)
            try:
                return Binary(fpath)
            except OSError as ex:
                source.critical("can't open %s: %s", fpath, ex)
                return None

    register_function(FSPATH)



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""provide class to do Apache rewrite rules'job inside cubicweb (though functionnalities
are much more limited for the moment)

"""



from re import compile

from cubicweb.web import Redirect
from cubicweb.web.component import Component

class RewriteCond(object):
    def __init__(self, condition, match='host', rules=(), action='rewrite'):
        self.condition = compile(condition)
        assert match in ('host', 'path'), match
        self.match_part = match
        self.rules = []
        for rule, replace in rules:
            rulergx = compile(rule)
            self.rules.append( (rulergx, replace) )
        assert action in ('rewrite', 'redirect', 'stop'), action
        self.process = getattr(self, 'action_%s' % action)

    def match(self, **kwargs):
        self._match = self.condition.match(kwargs[self.match_part])
        return not self._match is None

    def action_rewrite(self, path):
        for rgx, replace in self.rules:
            if not rgx.match(path) is None:
                matchdict = self._match.groupdict() or None
                if not matchdict is None:
                    replace = replace % matchdict
                return rgx.sub(replace, path)
        return path

    def action_redirect(self, path):
        url = self.action_rewrite(path)
        raise Redirect(url)

    def action_stop(self, path):
        return path


class ApacheURLRewrite(Component):
    """inherit from this class with actual rules to activate apache style rewriting

    rules should have the form :

    [('condition pattern 1', [('rule1 pattern', 'replace expression'),
                              ('rule2 pattern', 'replace expression')],
     ('condition pattern 2', [('rule1 pattern', 'replace expression'),
                              ('rule2 pattern', 'replace expression')]
    ]

    for instance the equivalent of the following apache rules:

        RewriteCond %{HTTP_HOST} ^logilab\.fr
        RewriteRule ^/(.*) http://www.logilab.fr/$1 [L,R=301]

        RewriteCond %{HTTP_HOST} ^www\.logilab\.fr
        RewriteRule ^/(.*) http://localhost:8080/$1 [L,P]

        RewriteCond %{HTTP_HOST} ^(.+)\.logilab\.fr
        RewriteRule ^/(data/.*) http://localhost:8080/$1 [L,P]
        RewriteRule ^/(json.*) http://localhost:8080/$1 [L,P]
        RewriteRule ^/(.*) http://localhost:8080/m_%1/$1 [L,P]

    could be written (considering that no "host rewritting" is necessary):

      class MyAppRules(ApacheURLRewrite):
        rules = [
          RewriteCond('logilab\.fr', match='host',
                      rules=[('/(.*)', r'http://www.logilab.fr/\1')],
                      action='redirect'),
          RewriteCond('(www)\.logilab\.fr', match='host', action='stop'),
          RewriteCond('/(data|json)/', match='path', action='stop'),
          RewriteCond('(?P<cat>.*)\.logilab\.fr', match='host',
                      rules=[('/(.*)', r'/m_%(cat)s/\1')]),
        ]
    """
    __abstract__ = True
    __regid__ = 'urlrewriter'
    rules = []

    def get_rules(self, req):
        return self.rules

    def rewrite(self, host, path, req):
        for cond in self.get_rules(req):
            if cond.match(host=host, path=path):
                return cond.process(path)
        return path
